﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Threading.Tasks;

namespace Chapter08
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void Parse()
        {
            Console.Write("What’s your age? ");
            string age = Console.ReadLine();
            try
            {
                int yearOfBirth = DateTime.Now.Year - int.Parse(age);
                Console.WriteLine("I bet you were born around " + yearOfBirth + ".");
            }
            catch (FormatException)
            {
                Console.WriteLine("Invalid age specified.");
            }
        }

        static void TryParse()
        {
            Console.Write("What’s your age? ");
            string ageString = Console.ReadLine();
            int age;
            if (int.TryParse(ageString, out age))
            {
                int yearOfBirth = DateTime.Now.Year - age;
                Console.WriteLine("I bet you were born around " + yearOfBirth + ".");
            }
        }

        static void StackOverflow()
        {
            StackOverflow();
        }

        static void FileNotFound()
        {
            PrintFile(@"Z:\Wait\Where\What.txt");
        }

        static void PrintFile(string path)
        {
            foreach (string line in File.ReadAllLines(path))
            {
                Console.WriteLine(line);
            }
        }

        // Use this as the entry-point (rename to Main) to run our "type" command clone.
        static void Main2(string[] args)
        {
            if (args.Length != 1)
            {
                string executable = Environment.GetCommandLineArgs()[0];
                string executableFileName = Path.GetFileName(executable);
                Console.WriteLine("Usage: {0} <file>", executableFileName);
                return;
            }

            string path = args[0];
            try
            {
                PrintFile(path);
            }
            catch (FileNotFoundException ex)
            {
                Console.ForegroundColor = ConsoleColor.Red;
                Console.WriteLine(ex.Message);
                Console.ResetColor();
            }
        }

        static void ThrowingExceptions()
        {
            DisplayUserProfile("");
        }

        static void DisplayUserProfile(string user)
        {
            if (user == null)
                throw new ArgumentNullException("user");
            if (user == "")
                throw new ArgumentException("exception");
            if (!Profiles.Exists(user))
                throw new UserNotFoundException(user);
            // Do real work here.
        }

        static void HandingExceptions(string path)
        {
            try
            {
                PrintFile(path);
            }
            catch (FileNotFoundException ex)
            {
                PrintException(ex);
            }
            catch (UnauthorizedAccessException ex)
            {
                PrintException(ex);
            }
        }

        private static void PrintException(Exception ex)
        {
            Console.WriteLine(ex.Message);
        }

        static void PrintFile_Flawed(string path)
        {
            FileStream fs = File.OpenRead(path);
            StreamReader sr = new StreamReader(fs);
            string line;
            while ((line = sr.ReadLine()) != null)
            {
                Console.WriteLine(line);
            }
        }

        static void PrintFile_Finally(string path)
        {
            FileStream fs = File.OpenRead(path);
            try
            {
                StreamReader sr = new StreamReader(fs);
                try
                {
                    string line;
                    while ((line = sr.ReadLine()) != null)
                    {
                        Console.WriteLine(line);
                    }
                }
                finally
                {
                    sr.Close();
                }
            }
            finally
            {
                fs.Close();
            }
        }

        static void PrintFile_Using(string path)
        {
            using (FileStream fs = File.OpenRead(path))
            using (StreamReader sr = new StreamReader(fs))
            {
                string line;
                while ((line = sr.ReadLine()) != null)
                {
                    Console.WriteLine(line);
                }
            }
        }

        static void Disposable()
        {
            using (var resource = new MyResource())
            {
                resource.Write("Hello");
            } // Calls Dispose here
        }

        static void Locks_TheProblem()
        {
            const int N = 100000000;

            var count = new Counter();
            Parallel.Invoke(
                () => { for (int i = 0; i < N; i++) count.Increment(); },
                () => { for (int i = 0; i < N; i++) count.Decrement(); });

            Console.WriteLine(count.Value); // Do you expect N?
        }

        static void Locks_TheSolution()
        {
            const int N = 100000000;

            var count = new CounterWithLocks(); // Notice the change!
            Parallel.Invoke(
                () => { for (int i = 0; i < N; i++) count.Increment(); },
                () => { for (int i = 0; i < N; i++) count.Decrement(); });

            Console.WriteLine(count.Value); // Do you expect N?
        }
    }

    static class Profiles
    {
        public static bool Exists(string user)
        {
            // For the sake of demonstration.
            return false;
        }
    }

    class UserNotFoundException : Exception
    {
        public UserNotFoundException(string user)
        {
            User = user;
        }

        public string User { get; private set; }
    }

    class MyResource : IDisposable
    {
        private bool _disposed;

        public void Dispose()
        {
            Dispose(true /* called by user directly */);
            GC.SuppressFinalize(this); // Tell the GC not to finalize this object.
        }

        void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                // Clean up logic goes here.
                // Code can use "disposing" to distinguish between cases.
                _disposed = true;
            }
        }

        public void Write(string message)
        {
            if (_disposed)
                throw new ObjectDisposedException("MyResource" /* name of the resource */);
            // Do work here.
        }

        ~MyResource()
        {
            Dispose(false /* not called by user directly */);
        }
    }

    class Counter
    {
        private int _value;

        public void Increment()
        {
            _value++;
        }
        public void Decrement()
        {
            _value--;
        }

        public int Value { get { return _value; } }
    }

    class CounterWithLocks
    {
        private int _value;
        private object _lock = new object();

        public void Increment()
        {
            lock (_lock)
            {
                _value++;
            }
        }
        public void Decrement()
        {
            lock (_lock)
            {
                _value--;
            }
        }

        public int Value { get { return _value; } }
    }
}
