﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace Chapter11
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void Fields()
        {
            var bart = new Person("Bart", 26);
            Console.WriteLine(bart.ID);

            Universe.PrintStatistics();
        }

        static void Constants()
        {
            // Uses Math.PI
            Console.WriteLine(new Circle(2.0).Surface());

            // Look at ILDASM to see constants getting copied in
            Console.WriteLine(Naturals.TWO);
        }

        static void Enums()
        {
            // Printing enums
            {
                uint value = 0xFF0000;
                Color c = (Color)value;
                Console.WriteLine(c.ToString("G")); // "G" is the default and can be omitted.
                Console.WriteLine(c.ToString("D"));
                Console.WriteLine(c.ToString("X"));

                Console.WriteLine(Enum.Format(typeof(Color), value, "D"));
                Console.WriteLine(Enum.Format(typeof(Color), value, "X"));
                Console.WriteLine(Enum.Format(typeof(Color), value, "G"));
            }

            // Reflecting on enum values
            {
                foreach (string member in Enum.GetNames(typeof(Color)))
                    Console.WriteLine(member);

                foreach (uint member in Enum.GetValues(typeof(Color)))
                    Console.WriteLine(member);
            }

            // Using LINQ
            {
                var res = Enum.GetNames(typeof(Color)).Zip(
                    Enum.GetValues(typeof(Color)).Cast<uint>(),
                        (name, value) => new { Name = name, Value = value }
                    );

                foreach (var member in res)
                    Console.WriteLine(member);
            }

            // Checking defined enum values
            {
                PrintColor(0xFF0000);
                PrintColor(0xFF0001); // Oops...
            }

            // Looking up enum values by name
            {
                Console.Write("Enter a color name: ");
                var colorName = Console.ReadLine();
                Color color;
                if (!Enum.TryParse(colorName, /*ignoreCase*/ true, out color))
                    Console.WriteLine("Color {0} doesn’t exist.", colorName);
                else
                    Console.WriteLine(color);
            }
        }

        static void PrintColor(uint color)
        {
            if (!Enum.IsDefined(typeof(Color), color))
                throw new ArgumentOutOfRangeException();
            Color c = (Color)color;
            Console.WriteLine(c);
        }

        static void FlagsEnums()
        {
            Console.Write("Enter FileShare flags: ");
            var input = Console.ReadLine();

            FileShare fileShare;
            if (!Enum.TryParse(input, /*ignoreCase*/ true, out fileShare))
            {
                Console.WriteLine("Invalid flags specified.");
                return;
            }
            if ((fileShare & FileShare.Read) == FileShare.Read)
                Console.WriteLine("Read flag set.");
        }

        static void Properties()
        {
            var person = new Person("Bart", 26);
            Console.WriteLine(person.Name + " is " + person.Age);
        }

        static void Indexers()
        {
            // Indexer getters and setters
            {
                var phoneBook = new Dictionary<string, int>();
                phoneBook["Bart"] = 51662;
                phoneBook["John"] = 96207;

                Console.WriteLine("Bart’s phone number is: " + phoneBook["Bart"]);
            }

            // Collection initializers
            {
                var phoneBook = new Dictionary<string, int>
                {
                    { "Bart", 51662 },
                    { "John", 96207 },
                };

                Console.WriteLine("Bart’s phone number is: " + phoneBook["Bart"]);
            }

            // Defining indexers
            {
                Vector victor = new Vector(1, 2, 3);

                Console.WriteLine("X = " + victor[0]);

                // We’re leaving the implementation of an indexer with a string-typed parameter
                // as an exercise for the reader.
                // Console.WriteLine("Y = " + victor["y"]);

                Console.WriteLine("Z = " + victor[2]);
            }
        }
    }

    class Person
    {
        private static int s_population;

        private readonly int _id;
        private string _name;
        private int _age;

        public Person(string name, int age)
        {
            _id = s_population++;

            _name = name;
            _age = age;
        }

        public int ID
        {
            get { return _id; }
        }

        public string Name
        {
            get { return _name; }
            set
            {
                if (string.IsNullOrEmpty(value))
                    throw new ArgumentNullException("value");

                _name = value;
            }
        }

        public int Age
        {
            get { return _age; }
            set { _age = value; }
        }

        public static int Population
        {
            get { return s_population; }
        }

        public static bool IsYounger(Person x, Person y)
        {
            // Don’t forget to add null-checks!
            return x._age < y._age;
        }

        public override string ToString()
        {
            return _name + " is " + this._age;
        }
    }

    class Universe
    {
        public static void PrintStatistics()
        {
            Console.WriteLine("Number of citizens: {0}", Person.Population);
        }
    }

    class Circle
    {
        private double _radius;

        public Circle(double radius)
        {
            _radius = radius;
        }

        public double Surface()
        {
            return Math.PI * _radius * _radius;
        }
    }

    class Naturals
    {
        public const int ZERO = 0;
        public const int ONE = ZERO + 1;
        public const int TWO = ONE + ONE;
    }

    enum Color : uint
    {
        Red = 0xFF0000,
        Green = 0x00FF00,
        Blue = 0x0000FF,
    }

    class Vector
    {
        private int _x, _y, _z;

        public Vector(int x, int y, int z)
        {
            _x = x; _y = y; _z = z;
        }

        public int this[int index]
        {
            get
            {
                if (index == 0) return _x;
                else if (index == 1) return _y;
                else if (index == 2) return _z;
                else throw new ArgumentOutOfRangeException();
            }
        }

        // Insert the implementation of an indexer with a string-typed parameter
        // here, allowing the user to write things like vector["x"] or vector["Y"].
    }
}
