﻿using System;

namespace Chapter14
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void Encapsulate()
        {
            Counter c1 = new Counter();
            Counter c2 = new Counter();

            c1.Increment(); // 1
            c2.Decrement(); // -1
            c2.Increment(); // 0
            c1.Increment(); // 2
            c1.Decrement(); // 1
            c2.Increment(); // 1

            Console.WriteLine(c1.Count);
            Console.WriteLine(c2.Count);
        }

        static void Polymorphism()
        {
            Counter counter = new NonNegativeCounter();
            counter.Increment();
            counter.Decrement();
            counter.Decrement(); // Will throw an exception.
        }

        static void Inheritance()
        {
            var student = new Student("Bart", 27);
            Console.WriteLine(student.Name + " is " + student.Age);
        }

        static void Virtual()
        {
            Person bart = new Student("Bart", 27, "C# Academy");
            Person lisa = new Student("Lisa", 23, "VB Academy");
            Console.WriteLine(bart.ToString()); // OO.Student (the full type name)
            Console.WriteLine(bart.Equals(lisa)); // false (no reference equality)
            Console.WriteLine(lisa.GetHashCode()); // Some 32-bit integer value
            Console.WriteLine(lisa.GetType()); // OO.Student (ToString of Type object)
        }

        static void Abstract_Shapes()
        {
            Shape s1 = new Circle(2.0);
            Console.WriteLine(s1.Area);
            Shape s2 = new Square(3.0);
            Console.WriteLine(s2.Circumference);
        }

        static void Interfaces()
        {
            var qux = new Qux();

            // Implicit implementation
            qux.Foo();
            
            // Explicit implementation
            ((IBar)qux).Foo();
        }
    }

    class Counter : IComparable<Counter>
    {
        private int _id, _i;
        private static int s_n;

        public Counter()
        {
            _id = s_n++;
        }

        public virtual int Increment()
        {
            Console.WriteLine("{0} -> {1}", _id, ++_i);
            return _i;
        }

        public virtual int Decrement()
        {
            Console.WriteLine("{0} -> {1}", _id, --_i);
            return _i;
        }

        public int Count
        {
            get { return _i; }
            protected set { _i = value; }
        }

        public int CompareTo(Counter other)
        {
            // Negative if "this < other", positive if "this > other", zero if equal
            return this.Count - other.Count;
        }
    }

    class Tracking
    {
        private Counter _activeVisitors = new Counter();

        public User Logon(string user, string password)
        {
            // Authenticate and authorize. Throw on failure.

            _activeVisitors.Increment();

            return new User(user);
        }

        public void Logoff(User user)
        {
            // Some plumbing...
            _activeVisitors.Decrement();
        }

        public int CurrentVisitorCount
        {
            get { return _activeVisitors.Count; }
        }
    }

    class User
    {
        private string user;

        public User(string user)
        {
            // TODO: Complete member initialization
            this.user = user;
        }
    }

    class ResettableCounter : Counter
    {
        public void Reset()
        {
            if (Count < 0)
                while (Count < 0)
                    Increment();
            if (Count > 0)
                while (Count > 0)
                    Decrement();
        }
    }

    class NonNegativeCounter : Counter
    {
        public override int Decrement()
        {
            if (Count == 0)
                throw new InvalidOperationException("Can’t count below zero.");

            return base.Decrement();
        }
    }

    class StepCounter : Counter
    {
        private int _step;

        public StepCounter(int step)
        {
            _step = step;
        }

        public override int Increment()
        {
            Count += _step; // Remember the setter is protected!
            return Count;
        }

        public override int Decrement()
        {
            Count -= _step; // Remember the setter is protected!
            return Count;
        }
    }

    class Person
    {
        public Person(string name, int age)
        {
            Name = name;
            Age = age;
        }

        public string Name { get; private set; }
        public int Age { get; private set; }
    }

    sealed class Student : Person
    {
        public Student(string name, int age)
            : base(name, age)
        { }

        public Student(string name, int age, string college)
            : this(name, age)
        {
            College = college;
        }

        public string College { get; private set; }

        public override string ToString()
        {
            return base.Name + " is " + base.Age + " and studies at " + this.College;
        }
    }

    class Bar
    {
        public void Foo() { Console.WriteLine("Bar.Foo"); }
        public static void Qux() { Console.WriteLine("Bar::Qux"); }
    }

    // Will produce warnings about hiding
    class Baz : Bar
    {
        public void Foo() { Console.WriteLine("Baz.Foo"); }
        public static void Qux() { Console.WriteLine("Baz::Qux"); }
    }

    class Baz2 : Bar
    {
        public new void Foo() { Console.WriteLine("Baz2.Foo"); }
        public new static void Qux() { Console.WriteLine("Baz2::Qux"); }
    }

    abstract class Shape
    {
        public virtual double Area
        {
            get { return 0.0; }
        }

        public virtual double Circumference
        {
            get { return 0.0; }
        }
    }

    class Circle : Shape
    {
        public Circle(double radius)
        {
            Radius = radius;
        }

        public double Radius { get; private set; }

        public sealed override double Area
        {
            get { return Math.PI * Radius * Radius; }
        }

        public sealed override double Circumference
        {
            get { return Math.PI * 2 * Radius; }
        }
    }

    class Square : Shape
    {
        public Square(double side)
        {
            Side = side;
        }

        public double Side { get; private set; }

        public sealed override double Area
        {
            get { return Side * Side; }
        }

        public sealed override double Circumference
        {
            get { return 4 * Side; }
        }
    }

    interface IBar
    {
        void Foo();
    }

    interface IBaz
    {
        void Foo();
    }

    class Qux : IBar, IBaz
    {
        void IBar.Foo()
        {
            Console.WriteLine("IBar.Foo");
        }

        public void Foo()
        {
            Console.WriteLine("IBaz.Foo");
        }
    }
}
