﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;

namespace Chapter15
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void Lists()
        {
            List<int> xs = new List<int> { 1, 2, 3 };
            List<string> names = new List<string> { "Bart", "John" };

            foreach (int x in xs)
                Console.WriteLine(x);

            foreach (string name in names)
                Console.WriteLine(name);
        }

        static void Dictionaries()
        {
            var numbers = new Dictionary<string, int> {
                { "Bart", 911 },
                { "John", 119 }
            };

            foreach (KeyValuePair<string, int> entry in numbers)
            {
                Console.WriteLine(entry.Key + " -> " + entry.Value);
            }
        }

        static void Performance()
        {
            const int N = 10000000; // 10 million elements
            var array = new int[N];
            var arrList = new ArrayList(N);
            var genList = new List<int>(N);

            Measure("Write ArrayList: {0}", () =>
            {
                for (int i = 0; i < N; i++)
                    arrList.Add(1); // Expected: BOXING!
            });

            Measure("Read ArrayList: {0}", () =>
            {
                int sum = 0;
                for (int i = 0; i < N; i++)
                    sum += (int)arrList[i]; // Expected: UNBOXING!
            });

            PrintGCStats();

            Measure("Write List<int>: {0}", () =>
            {
                for (int i = 0; i < N; i++)
                    genList.Add(1); // Expected: capacity check
            });

            Measure("Read List<int>: {0}", () =>
            {
                int sum = 0;
                for (int i = 0; i < N; i++)
                    sum += genList[i];
            });

            PrintGCStats();

            Measure("Write int array: {0}", () =>
            {
                for (int i = 0; i < N; i++)
                    array[i] = 1;
            });

            Measure("Read int array: {0}", () =>
            {
                int sum = 0;
                for (int i = 0; i < N; i++)
                    sum += array[i];
            });

            PrintGCStats();
        }

        static void Measure(string message, Action a)
        {
            var sw = new Stopwatch();
            sw.Start();
            a();
            sw.Stop();
            Console.WriteLine(message, sw.Elapsed);
        }

        static void PrintGCStats()
        {
            Console.WriteLine("Gen 0: {0}, Gen 1: {1}, Gen 2: {2}\n", GC.CollectionCount(1), GC.CollectionCount(2), GC.CollectionCount(3));
        }

        static void Default()
        {
            var cache = new Cache<int>(42);
            Console.WriteLine(cache.Value);

            cache.Reset();
            Console.WriteLine(cache.Value);
        }

        static void Constraints_BaseClass()
        {
            var appleMixer = new FruitMixer<Apple>(new Jonagold());
        }

        static void Constraints_ValueClass()
        {
            var hasRef = new HasRef<string>("Hello");
            var hasVal = new HasVal<int>(42);

            // Doesn't compile
            //var hasRef = new HasRef<int>(42);
            //var hasVal = new HasVal<string>("Hello");
        }

        static void GenericMethods_Explicit()
        {
            IEnumerable<int> evens = Enumerable.Where<int>(
                new int[] { 1, 2, 3 },
                (int x) => x % 2 == 0
            );

            foreach (var even in evens)
                Console.WriteLine(even);
        }

        static void GenericMethods_Inferred()
        {
            var evens = Enumerable.Where(new[] { 1, 2, 3 }, x => x % 2 == 0);

            foreach (var even in evens)
                Console.WriteLine(even);
        }

        static void Covariance_BrokenArrays()
        {
            var apple1 = new Jonagold();
            var apple2 = new Jonagold();
            var apple3 = new Jonagold();

            Apple[] apples = new Apple[] { apple1, apple2, apple3 };

            // Because of array type covariance we can write the following.
            Fruit[] fruits = apples;

            // We’re putting a Coconut, which is a Fruit, in a Fruit array. This is fine.
            // Or not? This will throw an ArrayTypeMismatchException at runtime!
            fruits[2] = new Coconut();
        }

        static void Covariance_SafeInterfaces()
        {
            var apple1 = new Jonagold();
            var apple2 = new Jonagold();
            var apple3 = new Jonagold();

            IEnumerable<Apple> apples = new Apple[] { apple1, apple2, apple3 };

            // Because of IEnumerable<T>’s covariance we can write the following.
            // Make sure to compile using C# 4.0 to have this work.
            IEnumerable<Fruit> fruits = apples;

            // There is no way to put a Coconut in an IEnumerable<Fruit>, hence this does
            // not work anymore: fruits[2] = new Coconut();
        }

        static void Contravariance()
        {
            IComparer<Fruit> compareFruits = new CompareFruitBySize();
            IComparer<Apple> compareApples = compareFruits;

            // Just the fact the assignment above works, reveals contravariance.
        }

        class CompareFruitBySize : IComparer<Fruit>
        {
            public int Compare(Fruit x, Fruit y)
            {
                // Left for reader: implement based on (added) Fruit properties.
                throw new NotImplementedException();
            }
        }

    }

    class Cache<T>
    {
        private T _value;

        public Cache(T value)
        {
            _value = value;
        }

        public T Value
        {
            get
            {
                return _value;
            }
        }

        // Provide a way to recompute the value, for example, using a delegate.
        // In fact, the BCL has a similar construct called Lazy<T>.

        public void Reset()
        {
            _value = default(T);
        }
    }

    class FruitMixer<T> where T : Fruit
    {
        private T[] _ingredients;

        public FruitMixer(params T[] ingredients)
        {
            _ingredients = ingredients;
        }
    }

    class Fruit
    {
    }

    class Apple : Fruit
    {
    }

    class Jonagold : Apple
    {
    }

    class Coconut : Fruit
    {
    }

    class ServicePool<T> : IService // expose the service to the outside
        where T : IService, // need to be able to delegate to a worker
                  new() // need to be able to instantiate workers as needed
    {
        public ServicePool(int min, int max) { /* TODO */ }

        public void TransferFunds(string sourceAccount, string targetAccount, decimal amount)
        {
            // Forward to some service in the pool
        }
    }

    // Any IService interface would do, so we just give a simple example here.
    interface IService
    {
        void TransferFunds(string sourceAccount, string targetAccount, decimal amount);
    }

    class HasRef<T> where T : class
    {
        private T _reference;

        public HasRef(T reference)
        {
            _reference = reference;
        }

        public T Reference { get { return _reference; } }
    }

    class HasVal<T> where T : struct
    {
        private T _value;

        public HasVal(T value)
        {
            _value = value;
        }

        public T Value { get { return _value; } }
    }
}
