﻿using System;
using System.ComponentModel;
using System.Threading;
using System.Windows.Forms;

namespace Chapter18
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void CountDown_Events()
        {
            var countDown = new CountDown(5);

            // Hook up event handlers
            countDown.Tick += currentSecond => Console.WriteLine(currentSecond);
            countDown.Finished += () => Console.Beep();
            countDown.Start();
        }

        static void CountDown_Events_Unhook()
        {
            var countDown = new CountDown(5);

            // Hook up event handlers
            Action<uint> tickHandler = currentSecond => Console.WriteLine(currentSecond);
            Action finishedHandler = null; // now the variable is pre-assigned!
            finishedHandler = () =>
            {
                Console.Beep();
                // Finished has been called: we can unhook both Tick and Finished now.
                countDown.Tick -= tickHandler;
                countDown.Finished -= finishedHandler;
            };

            countDown.Start();
        }

        static void CountDown_Events_HandlersAsMethods()
        {
            ProgramWithCountDown.Main2();
        }

        static void CountDown_Events_Pattern()
        {
            ProgramWithCountDown2.Main2();
        }

        static void CountDown_Events_Inheritance()
        {
            var countDown = new FancyCountDown(5);
            countDown.Tick += (sender, args) => Console.WriteLine(args.Seconds);
            countDown.Finished += (sender, args) => Console.Beep();

            countDown.Start();

            Thread.Sleep(2000);

            countDown.EmergencyStop(); // Triggers Finished event
        }
    }

    class ProgramWithCountDown
    {
        private static CountDown s_countDown;

        public static void Main2()
        {
            s_countDown = new CountDown(5);
            s_countDown.Tick += TickHandler;
            s_countDown.Finished += FinishedHandler;
            s_countDown.Start();
        }

        static void TickHandler(uint currentSecond)
        {
            Console.WriteLine(currentSecond);
        }

        static void FinishedHandler()
        {
            Console.Beep();
            s_countDown.Tick -= TickHandler;
            s_countDown.Finished -= FinishedHandler;
        }

        static void NotifyProperty_UI()
        {
            Application.Run(new PersonEditor());
        }
    }

    class CountDown
    {
        private uint _seconds;

        public CountDown(uint seconds)
        {
            _seconds = seconds;
        }

        public void Start()
        {
            new Thread(() =>
            {
                uint n = _seconds;
                while (n > 0u)
                {
                    var tick = Tick;
                    if (tick != null)
                        tick(n);

                    Thread.Sleep(1000);
                    n--;
                }

                var finished = Finished;
                if (finished != null)
                    finished();
            }).Start();
        }

        public event Action<uint> Tick;
        public event Action Finished;
    }

    class ProgramWithCountDown2
    {
        private static CountDown2 s_countDown;

        public static void Main2()
        {
            s_countDown = new CountDown2(5);
            s_countDown.Tick += TickHandler;
            s_countDown.Finished += FinishedHandler;
            s_countDown.Start();
        }

        static void TickHandler(object sender, TickEventArgs e)
        {
            Console.WriteLine(e.Seconds);
        }

        static void FinishedHandler(object sender, EventArgs e)
        {
            Console.Beep();
            // Finished has been called: we can unhook both Tick and Finished now.
            s_countDown.Tick -= TickHandler;
            s_countDown.Finished -= FinishedHandler;
        }
    }

    class CountDown2
    {
        private uint _seconds;

        public CountDown2(uint seconds)
        {
            _seconds = seconds;
        }

        public void Start()
        {
            new Thread(() =>
            {
                uint n = _seconds;
                while (n > 0u)
                {
                    var tick = Tick;
                    if (tick != null)
                        tick(this, new TickEventArgs(n));

                    Thread.Sleep(1000);
                    n--;
                }

                var finished = Finished;
                if (finished != null)
                    finished(this, EventArgs.Empty);
            }).Start();
        }

        public event EventHandler<TickEventArgs> Tick;
        public event EventHandler Finished;
    }

    class TickEventArgs : EventArgs
    {
        public TickEventArgs(uint seconds)
        {
            this.Seconds = seconds;
        }

        public uint Seconds { get; private set; }
    }

    class CountDown3
    {
        private uint _seconds;

        public CountDown3(uint seconds)
        {
            _seconds = seconds;
        }

        public void Start()
        {
            new Thread(() =>
            {
                uint n = _seconds;
                while (n > 0u)
                {
                    OnTick(new TickEventArgs(n));

                    Thread.Sleep(1000);
                    n--;
                }

                OnFinished(EventArgs.Empty);
            }).Start();
        }

        public event EventHandler<TickEventArgs> Tick;
        public event EventHandler Finished;

        protected virtual void OnTick(TickEventArgs e)
        {
            var tick = Tick;
            if (tick != null)
            {
                tick(this, e);
            }
        }

        protected virtual void OnFinished(EventArgs e)
        {
            var finished = Finished;
            if (finished != null)
            {
                finished(this, e);
            }
        }
    }

    class FancyCountDown : CountDown3
    {
        public FancyCountDown(uint seconds)
            : base(seconds)
        {
        }

        public void EmergencyStop()
        {
            OnFinished(EventArgs.Empty);
        }
    }

    class Person : INotifyPropertyChanging, INotifyPropertyChanged
    {
        private string _name;
        private int _age;

        public Person(string name, int age)
        {
            _name = name;
            _age = age;
        }

        public event PropertyChangingEventHandler PropertyChanging;
        public event PropertyChangedEventHandler PropertyChanged;

        public string Name
        {
            get { return _name; }
            set
            {
                if (_name != value)
                {
                    OnPropertyChanging("Name");
                    _name = value;
                    OnPropertyChanged("Name");
                }
            }
        }

        public int Age
        {
            get { return _age; }
            set
            {
                if (_age != value)
                {
                    OnPropertyChanging("Age");
                    _age = value;
                    OnPropertyChanged("Age");
                }
            }
        }

        protected virtual void OnPropertyChanging(string propertyName)
        {
            var propertyChanging = PropertyChanging;
            if (propertyChanging != null)
            {
                propertyChanging(this, new PropertyChangingEventArgs(propertyName));
            }
        }

        protected virtual void OnPropertyChanged(string propertyName)
        {
            var propertyChanged = PropertyChanged;
            if (propertyChanged != null)
            {
                propertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }
    }
}
