﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;

namespace Chapter20
{
    static class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void ListOfOperators()
        {
            var queryOperators = from m in typeof(Enumerable).GetMethods()
                                 where m.IsStatic
                                 group m by m.Name into queryOperator
                                 orderby queryOperator.Key /* Key is method name m.Name */
                                 select new
                                 {
                                     Name = queryOperator.Key,
                                     Overloads = queryOperator.Count()
                                 };

            foreach (var queryOperator in queryOperators)
                Console.WriteLine("{0} has {1} overload(s)", queryOperator.Name, queryOperator.Overloads);
        }

        static void Iterators()
        {
            foreach (var x in GetIntegers().Take(10))
                Console.WriteLine(x);

            var e = CountDown();
            while (e.MoveNext())
                Console.WriteLine(e.Current);
        }

        static IEnumerable<int> GetIntegers()
        {
            int n = 0;
            while (true)
                yield return n++;
        }

        static IEnumerator<int> CountDown()
        {
            Console.WriteLine("Before three");
            yield return 3;
            Console.WriteLine("After three");
            Console.WriteLine("Before two");
            yield return 2;
            Console.WriteLine("After two");
            Console.WriteLine("Before one");
            yield return 1;
            Console.WriteLine("After one");
            Console.WriteLine("Before zero");
            yield return 0;
        }

        static void QueryOperatorsUsingIterators()
        {
            // First example of our Filter iterator.
            {
                var evens = Filter(new[] { 2, 5, 7, 4, 1, 9, 5, 8 }, n =>
                {
                    Console.WriteLine("Filtering " + n);
                    return n % 2 == 0;
                });

                foreach (int n in evens)
                    Console.WriteLine(n);
            }

            // Using Filter multiple times.
            {
                var multiplesOfSix = Filter(
                    Filter(
                        new[] { 1, 2, 3, 4, 5, 6, 7, 8, 9 },
                        n => n % 2 == 0
                    ),
                    n => n % 3 != 0
                );

                foreach (var n in multiplesOfSix)
                    Console.WriteLine(n);
            }

            // Using LINQ extension methods, it looks like this.
            {
                var multiplesOfSix = new[] { 1, 2, 3, 4, 5, 6, 7, 8, 9 }
                    .Where(n => n % 2 == 0)
                    .Where(n => n % 3 != 0);

                foreach (var n in multiplesOfSix)
                    Console.WriteLine(n);
            }
        }

        static IEnumerable<T> Filter<T>(IEnumerable<T> source, Func<T, bool> filter)
        {
            foreach (T item in source)
                if (filter(item))
                    yield return item;
        }

        static void LazyEvaluation()
        {
            // Query expression reuse.
            {
                var nums = new List<int> { 2, 5, 7, 4, 1, 9 };
                var evens = from n in nums
                            where n % 2 == 0
                            select n;

                // Shows 2, 4
                foreach (var n in evens)
                    Console.WriteLine(n);

                // Somehow the collection changes, e.g. because the user edits the data.
                nums.Add(5);
                nums.Add(8);

                // Shows 2, 4, 8 - no need to redeclare the query
                foreach (var n in evens)
                    Console.WriteLine(n);
            }

            // Exceptions come out lazily as well! 
            {
                var nums = new List<int> { 2, 5, 7, 4, 1, 9 };

                IEnumerable<int> inverse = null;
                try
                {
                    inverse = from n in nums
                              select 1 / n;
                }
                catch (DivideByZeroException)
                {
                    // Bug: you’ll never get here.
                }

                // But the following line can throw, since the query executes here!
                foreach (var n in inverse)
                    Console.WriteLine(n);
            }

            // Throwing exceptions for iterators is tricky. Filter2 does it right.
            {
                var res = Filter2<int>(null, x => true); // throws here

                foreach (var x in res)
                    Console.WriteLine(x);
            }
        }

        static IEnumerable<T> Filter2<T>(IEnumerable<T> source, Func<T, bool> filter)
        {
            if (source == null)
                throw new ArgumentNullException("source");
            if (filter == null)
                throw new ArgumentNullException("filter");

            return FilterInternal(source, filter);
        }

        static IEnumerable<T> FilterInternal<T>(IEnumerable<T> source, Func<T, bool> filter)
        {
            foreach (T item in source)
                if (filter(item))
                    yield return item;
        }

        static void IteratorsBehindTheScenes()
        {
            // Look at StartWith in ILDASM.

            var res =
                // { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 }
                StartWith(
                // { 1, 2, 3, 4, 5, 6, 7, 8, 9 }
                    StartWith(
                // { 2, 3, 4, 5, 6, 7, 8, 9 }
                        Enumerable.Range(2, 8),
                    1),
                0);

            foreach (var x in res)
                Console.WriteLine(x);
        }

        static IEnumerable<T> StartWith<T>(this IEnumerable<T> tail, T head)
        {
            yield return head;
            foreach (T item in tail)
                yield return item;
        }

        static void StandardQueryOperators()
        {
            // Range
            {
                IEnumerable<int> firstTen = Enumerable.Range(0, 10); // { 0 .. 9 }
                foreach (var x in firstTen)
                    Console.WriteLine(x);

                IEnumerable<char> aToz = Enumerable.Range(0, 26).Select(i => (char)('a' + i));
                foreach (var c in aToz)
                    Console.Write(c);
                Console.WriteLine();
            }

            // Repeat
            {
                IEnumerable<string> heyHeyHey = Enumerable.Repeat("Hey", 3);
                foreach (var hey in heyHeyHey)
                    Console.WriteLine(hey);
            }

            // Empty
            {
                IEnumerable<int> nothing = Enumerable.Empty<int>();
                foreach (var e in nothing)
                    Console.WriteLine(e);
            }

            // Return (non-standard)
            {
                IEnumerable<int> answer = Return(42);
                foreach (var x in answer)
                    Console.WriteLine(x);
            }

            // DefaultIfEmpty
            {
                var src1 = Enumerable.Empty<int>();

                var res1 = src1.DefaultIfEmpty(); // { 0 }
                foreach (var x in res1)
                    Console.WriteLine(x);

                var res2 = src1.DefaultIfEmpty(99); // { 99 }
                foreach (var x in res2)
                    Console.WriteLine(x);

                var src2 = Enumerable.Range(1, 9);

                var res3 = src2.DefaultIfEmpty(); // { 1 .. 9 }
                foreach (var x in res3)
                    Console.WriteLine(x);
            }

            // Where
            {
                // { “Bart”, “Lisa” }
                var res1 = new[] { "Bart", "Homer", "Lisa" }.Where(s => s.Length == 4);
                foreach (var name in res1)
                    Console.WriteLine(name);

                // { “Zero”, “Two” }
                var res2 = new[] { "Zero", "One", "Two", "Three" }.Where((s, i) => i % 2 == 0);
                foreach (var num in res2)
                    Console.WriteLine(num);
            }

            // Distinct
            {
                var nums = new[] { 5, 2, 1, 3, 1, 4, 5 }.Distinct(); // { 5, 2, 1, 3, 4 }
                foreach (var x in nums)
                    Console.WriteLine(x);
            }

            // Skip[While]
            {
                var res1 = Enumerable.Range(0, 10).Skip(5); // { 5, 6, 7, 8, 9 }
                foreach (var x in res1)
                    Console.WriteLine(x);

                var res2 = Enumerable.Range(0, 10).SkipWhile(n => n < 8); // { 8, 9 }
                foreach (var x in res2)
                    Console.WriteLine(x);
            }

            // Take[While]
            {
                var res1 = Enumerable.Range(0, 10).Take(5); // { 0, 1, 2, 3, 4 }
                foreach (var x in res1)
                    Console.WriteLine(x);

                var res2 = Enumerable.Range(0, 10).TakeWhile(n => n < 8); // { 0, 1, 2, 3, 4, 5, 6, 7 }
                foreach (var x in res2)
                    Console.WriteLine(x);
            }

            // First[OrDefault]
            {
                // Without predicates
                {
                    int first = new[] { 42, 24, 12, 123 }.First(); // 42
                    Console.WriteLine(first);

                    int zero = new int[0].FirstOrDefault();
                    Console.WriteLine(zero);
                }

                // With predicates
                {
                    int first = new[] { 42, 24, 12, 123 }.First(n => n < 30); // 24
                    Console.WriteLine(first);

                    int zero = new[] { 42, 24, 12, 123 }.FirstOrDefault(n => n > 999);
                    Console.WriteLine(zero);
                }
            }

            // Last[OrDefault]
            {
                // Without predicates
                {
                    int last = new[] { 42, 24, 12, 123 }.Last(); // 123
                    Console.WriteLine(last);

                    int zero = new int[0].LastOrDefault();
                    Console.WriteLine(zero);
                }

                // With predicates
                {
                    int last = new[] { 42, 24, 12, 123 }.Last(n => n < 30); // 12
                    Console.WriteLine(last);

                    int zero = new[] { 42, 24, 12, 123 }.LastOrDefault(n => n > 999);
                    Console.WriteLine(zero);
                }
            }

            // Single[OrDefault]
            {
                int oneAndOnly = new[] { 42 }.Single(); // 42
                Console.WriteLine(oneAndOnly);

                int zero = new[] { 1, 2, 3 }.SingleOrDefault(); // More than one element
                Console.WriteLine(zero);
            }

            // ElementAt
            {
                var two = new[] { "Zero", "One", "Two", "Three" }.ElementAt(2);
                Console.WriteLine(two);
            }

            // OfType
            {
                var src = new object[] { 42, "Bart", DateTime.Now, 123 };

                // { 42, 123 }
                IEnumerable<int> res1 = src.Where(o => o is int).Select(o => (int)o);
                foreach (var x in res1)
                    Console.WriteLine(x);

                // { 42, 123 }
                IEnumerable<int> res2 = src.OfType<int>();
                foreach (var x in res1)
                    Console.WriteLine(x);
            }

            // Select
            {
                IEnumerable<int> twice = Enumerable.Range(0, 10).Select(n => n * 2);
                foreach (var x in twice)
                    Console.WriteLine(x);

                IEnumerable<string> xs = Enumerable.Range(0, 10).Select(n => n.ToString());
                foreach (var s in xs)
                    Console.WriteLine(s);

                // { “0 is Zero”, “1 is One”, “2 is Two”, “3 is Three” }
                IEnumerable<string> numbers = new[] { "Zero", "One", "Two", "Three" }
                    .Select((s, i) => i + " is " + s);
                foreach (var n in numbers)
                    Console.WriteLine(n);
            }

            // Cast
            {
                var res1 = from int x in new ArrayList { 1, 2, 3 }
                           select x;
                foreach (var x in res1)
                    Console.WriteLine(x);

                var res2 = from x in new ArrayList { 1, 2, 3 }.Cast<int>()
                           select x;
                foreach (var x in res2)
                    Console.WriteLine(x);
            }

            // SelectMany
            {
                /*
                 * Import the Northwind database using the LINQ to SQL designer.
                 * 
                var northwind = new NorthwindDataContext();
                
                var res = from product in northwind.Products
                          from supplier in product.Suppliers
                          select new { product.Name, supplier.City };

                foreach (var ps in res)
                    Console.WriteLine(ps.Name + " ships from " + ps.City);
                 */
            }

            // Zip
            {
                var res = Enumerable.Range(0, 26).Zip(
                    from i in Enumerable.Range(0, 26) select (char)(i + 'a'),
                    (n, c) => "The " + n + "th character of the Latin alphabet is " + c + "."
                );

                foreach (var c in res)
                    Console.WriteLine(c);
            }

            // OrderBy
            {
                // Same as from process in Process.GetProcesses()
                //         orderby process.ProcessName, process.WorkingSet64 descending
                //         select process
                var memoryHungry = Process.GetProcesses()
                    .OrderBy(process => process.ProcessName)
                    .ThenByDescending(process => process.WorkingSet64);

                foreach (var p in memoryHungry)
                    Console.WriteLine(p.ProcessName);


                // Same as from x in new[] { 5, 2, 6, 3, 4, 1 }
                //         orderby x
                //         select x
                var oneToSix = new[] { 5, 2, 6, 3, 4, 1 }.OrderBy(x => x); // { 1, 2, 3, 4, 5, 6 }

                foreach (var x in oneToSix)
                    Console.WriteLine(x);
            }

            // GroupBy
            {
                // Same as from x in new[] { 5, 2, 3, 5, 1, 1 }
                //         group x by x % 2 == 0
                //         ...
                {
                    var res = new[] { 5, 2, 3, 5, 1, 1 }
                        .GroupBy(x => x % 2 == 0);

                    foreach (var g in res)
                    {
                        Console.WriteLine(g.Key ? "Evens: " : "Odds: ");
                        foreach (int x in g)
                            Console.WriteLine("  " + x);
                    }
                }

                // Same as from x in new[] { 5, 2, 3, 5, 1, 1 }
                //         group x.ToString() by x % 2 == 0
                //         ...
                {
                    var res = new[] { 5, 2, 3, 5, 1, 1 }
                        .GroupBy(x => x % 2 == 0, x => x.ToString());

                    foreach (var g in res)
                    {
                        Console.WriteLine(g.Key ? "Evens: " : "Odds: ");
                        foreach (string x in g)
                            Console.WriteLine("  " + x);
                    }
                }
            }

            // GroupJoin
            {
                /*
                 * Import the Northwind database using the LINQ to SQL designer.
                 * 
                var northwind = new NorthwindDataContext();
                 
                // from product in northwind.Products
                // join order in northwind.Orders on product.ID equals order.Product into g
                // select product.Name + “ has “ + g.Count() + “ orders”
                var res = 
                    northwind.Products.GroupJoin(
                        northwind.Orders,
                        product => product.ID, // role of “primary key”
                        order => order.Product, // role of “foreign key”
                        (product, g) => product.Name + " has " + g.Count() + " orders"
                    );
                 */
            }

            // Join
            {
                /*
                 * Import the Northwind database using the LINQ to SQL designer.
                 * 
                var northwind = new NorthwindDataContext();
                 
                // from product in northwind.Products
                // join order in northwind.Orders on product.ID equals order.Product
                // select product.Name + “ is ordered by “ + order.Customer.Name
                var res = 
                    northwind.Products.Join(
                        northwind.Orders,
                        product => product.ID, // role of “primary key”
                        order => order.Product, // role of “foreign key”
                        (product, order) => product.Name + " is ordered by " + order.Customer.Name
                    );
                 */
            }

            // Aggregates
            {
                var res1 = new[] { 3, 7, 1, 5, 0, 8, 0, 4, 2 }
                    .Aggregate("", (s, n) => s + n.ToString());

                Console.WriteLine(res1); // "371508042"

                var res2 = new[] { 3, 7, 1, 5, 0, 8, 0, 4, 2 }
                    .Aggregate(new StringBuilder(), (sb, n) => sb.Append(n.ToString()));

                Console.WriteLine(res2); // "371508042"

                var res3 = new[] { 1, 2, 3 }
                    .Aggregate(
                        new { Count = 0, Sum = 0 },
                        (acc, n) => new { Count = acc.Count + 1, Sum = acc.Sum + n },
                        acc => acc.Sum / acc.Count /* integral division */
                    );

                Console.WriteLine(res3); // 2

                var procStats = from process in Process.GetProcesses()
                                group process by process.ProcessName into g
                                select new
                                {
                                    Process = g.Key,
                                    Count = g.Count(),
                                    Memory = g.Sum(p => p.WorkingSet64)
                                } into stat
                                where stat.Memory > 100 * 1024 * 1024
                                orderby stat.Memory descending
                                select stat;

                foreach (var stat in procStats)
                    Console.WriteLine(stat.Process + " has " + stat.Count + " instances, using a total of " + stat.Memory + "MB of memory");
            }

            // All and Any
            {
                var allEven = new[] { 2, 4, 6, 8, 10 }.All(x => x % 2 == 0); // true
                Console.WriteLine(allEven);

                var someOdd = new[] { 2, 3, 4, 5 }.Any(x => x % 2 == 1); // true
                Console.WriteLine(someOdd);

                var anyElement = new int[] { }.Any(); // false, sequence empty
                Console.WriteLine(anyElement);
            }

            // SequenceEqual
            {
                var f = new[] { 1, 2, 3 }.SequenceEqual(new[] { 1 }); // false
                Console.WriteLine(f);

                var s = new[] { 1, 2, 3 }.SequenceEqual(new[] { 2, 3, 1 }); // false
                Console.WriteLine(s);

                var t = new[] { 1, 2, 3 }.SequenceEqual(new[] { 1, 2, 3 }); // true
                Console.WriteLine(t);
            }

            // ToLookup
            {
                var procs = Process.GetProcesses().ToLookup(x => x.ProcessName);
                foreach (var instance in procs["notepad"])
                    Console.WriteLine(instance.Id);
            }
        }

        static IEnumerable<T> Return<T>(T element)
        {
            yield return element;
        }

        static void QueryPattern()
        {
            var query = from match in "Bart says \'Hallo\'"
                        where new Regex(".a..")
                        select match;

            foreach (var m in query)
                Console.WriteLine(m);
        }

        static void QueryingStrings()
        {
            var vowels = new[] { 'a', 'e', 'i', 'o', 'u' };
            var stats = from c in "Bart is sitting at Joey's in Bellevue"
                        where vowels.Contains(c)
                        group c by c into g
                        select new { Vowel = g.Key, Count = g.Count() };

            foreach (var stat in stats)
                Console.WriteLine(stat);
        }

        static void PLinq_Before()
        {
            // Watch CPU utilization in Task Manager.

            var res = from x in Enumerable.Range(0, 10000)
                      from y in Enumerable.Range(0, 10000)
                      where x > y
                      select x + y;

            var sw = new Stopwatch();
            sw.Start();
            res.ToList();
            sw.Stop();
            Console.WriteLine(sw.Elapsed);
        }

        static void PLinq_After()
        {
            // Watch CPU utilization in Task Manager.

            var res = from x in Enumerable.Range(0, 10000).AsParallel()
                      from y in Enumerable.Range(0, 10000)
                      where x > y
                      select x + y;

            var sw = new Stopwatch();
            sw.Start();
            res.ToList();
            sw.Stop();
            Console.WriteLine(sw.Elapsed);
        }

        static void PLinq_NoOrdering()
        {
            var res = from x in Enumerable.Range(0, 10000).AsParallel()
                      where x % 2 == 0
                      select x;

            foreach (var x in res)
                Console.WriteLine(x);
        }

        static void PLinq_Ordering()
        {
            var res = from x in Enumerable.Range(0, 10000).AsParallel().AsOrdered()
                      where x % 2 == 0
                      select x;

            foreach (var x in res)
                Console.WriteLine(x);
        }

        static void PLinq_ForAllMethod()
        {
            var res = from x in Enumerable.Range(0, 10).AsParallel()
                      where x % 2 == 0
                      select x;

            res.ForAll(x =>
                Console.WriteLine(Thread.CurrentThread.ManagedThreadId + " -> " + x)
            );
        }

        static void ExpressionTrees_HomoIconicity()
        {
            // Lambda as anonymous method.
            {
                var res1 = Calculate_Delegate(x => x + 1, 42);
                Console.WriteLine(res1); // 43
            }

            // Lambda as expression tree.
            {
                var res2 = Calculate_Expression(x => x + 1, 42);
                Console.WriteLine(res2); // 43
            }

            // Behind the scenes of expression trees.
            {
                var x = Expression.Parameter(typeof(int), "x");
                var f = Expression.Lambda<Func<int, int>>(
                    Expression.Add(
                        x,
                        Expression.Constant(1)
                    ),
                    x
                );

                var res3 = Calculate_Expression(f, 42);
                Console.WriteLine(res3); // 43
            }
        }

        static int Calculate_Delegate(Func<int, int> op, int arg)
        {
            return op(arg);
        }

        static int Calculate_Expression(Expression<Func<int, int>> op, int arg)
        {
            return op.Compile()(arg);
        }
    }

    public static class Extensions
    {
        public static IEnumerable<Match> Where(this string input, Func<object, Regex> regex)
        {
            return regex(null).Matches(input).Cast<Match>();
        }
    }
}
