﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Globalization;
using System.Numerics;
using System.Reflection;
using System.Text.RegularExpressions;

namespace Chapter26
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void ParsingStrings_Int()
        {
            string ageString;
            int age;
            do
            {
                Console.Write("Enter your age: ");
                ageString = Console.ReadLine();
            }
            while (!int.TryParse(ageString, out age));

            Console.WriteLine("Your age is {0}.", age);
        }

        static void ParsingStrings_Decimal()
        {
            Console.Write("Enter your culture: ");
            string culture = Console.ReadLine();
            string salaryString;
            decimal salary;
            do
            {
                Console.Write("Enter your salary: ");
                salaryString = Console.ReadLine();
            }
            while (!decimal.TryParse(salaryString, NumberStyles.Currency, new CultureInfo(culture), out salary));

            Console.WriteLine("Your salary is {0}.", salary);
        }

        static void SpecialValues()
        {
            Console.WriteLine("32-bit signed integers range from {0} to {1}.", int.MinValue, int.MaxValue);
        }

        static void Arrays()
        {
            // Rank
            {
                var xs = new int[10];
                var ys = new int[3, 4];
                var zs = new int[3][];

                Console.WriteLine(xs.Rank); // 1
                Console.WriteLine(ys.Rank); // 2
                Console.WriteLine(zs.Rank); // 1
            }

            // Array methods
            {
                var rnd = new Random();
                var xs = new int[20];
                for (int i = 0; i < 10; i++)
                    xs[i] = rnd.Next(10);

                Array.Resize(ref xs, 10); // drop last ten, like LINQ’s Take
                Array.Sort(xs); // in-place quicksort, like LINQ’s OrderBy
                Array.Reverse(xs); // in-place reversal, like LINQ’s Reverse

                var ys = new int[10];
                Array.Copy(xs, ys, xs.Length); // more overloads available
                Array.ForEach(ys, Console.WriteLine); // like a foreach-loop

                int fiveAt = Array.BinarySearch(ys, 5); // -1 if not found
                bool anyEven = Array.Exists(ys, i => i % 2 == 0); // like LINQ’s Any

                Console.WriteLine(fiveAt);
                Console.WriteLine(anyEven);

                // Other operations include:
                // - TrueForAll (like LINQ’s All operator)
                // - Find and FindAll (like LINQ’s Select operator)
                // - Ways to find indexes for an element
            }
        }

        static void ArraySegments()
        {
            var xs = new[] { 3, 7, 2, 5, 1, 4, 0, 9, 6, 8 };
            //                        ^^^^^^^^^^^^^
            var ys = new ArraySegment<int>(xs, 3, 5);

            // Retrieving start and end elements in the segment
            var fst = ys.Array[ys.Offset]; // 5
            var lst = ys.Array[ys.Offset + ys.Count - 1]; // 9

            Console.WriteLine(fst);
            Console.WriteLine(lst);
        }

        static void Buffers()
        {
            var xs = new[] { 3, 7, 2, 5, 1, 4, 0, 9, 6, 8 };
            var len1 = Buffer.ByteLength(xs); // xs.Length * sizeof(int)
            var len2 = xs.Length * sizeof(int);

            byte b1 = Buffer.GetByte(xs, 0); // 3 due to little-endian architecture
            byte b2 = Buffer.GetByte(xs, 1); // 0
            byte b3 = Buffer.GetByte(xs, 2); // 0
            byte b4 = Buffer.GetByte(xs, 3); // 0
            Console.WriteLine("{0} {1} {2} {3}", b1, b2, b3, b4);

            Console.WriteLine("0x{0:X}", xs[0]); // 0x00000003
            Buffer.SetByte(xs, 1, 0x42);         //       ^^
            Buffer.SetByte(xs, 3, 0x17);         //   ^^
            Console.WriteLine("0x{0:X}", xs[0]); // 0x17004203

            // 3 0 0 0 7 0 0 0 ...
            var ys = new byte[Buffer.ByteLength(xs)];
            Buffer.BlockCopy(xs, 0, ys, 0, xs.Length);
            Console.WriteLine(ys[0]); // 3
            Console.WriteLine(ys[3]); // 7
        }

        static void MathOps()
        {
            Console.WriteLine(Math.Sqrt(2)); // 1.4142135623731
            Console.WriteLine(Math.Log10(1000)); // 3
            Console.WriteLine(Math.Abs(-5)); // 5
            Console.WriteLine(Math.Sign(-5)); // -1
            Console.WriteLine(Math.Ceiling(Math.PI)); // 4
            Console.WriteLine(Math.Floor(Math.E)); // 2
            Console.WriteLine(Math.Round(Math.PI)); // 3
            Console.WriteLine(Math.Round(Math.E)); // 3
            Console.WriteLine(Math.Truncate(Math.PI)); // 3
            Console.WriteLine(Math.Truncate(Math.E)); // 2

            int x = int.MaxValue;
            int y = int.MaxValue;
            var z = Math.BigMul(x, y); // Now z is an Int64 (or long)
            Console.WriteLine(z);

            int num = 17;
            int den = 5;
            int rem;
            int div = Math.DivRem(num, den, out rem);
            Console.WriteLine("{0} * {1} + {2} = {3}", div, den, rem, num);

            double angle = 45 * Math.PI / 180;
            Console.WriteLine(Math.Sin(angle)); // Should be Math.Sqrt(2) / 2
            Console.WriteLine(Math.Cos(angle)); // Should be Math.Sqrt(2) / 2, too
            Console.WriteLine(Math.Tan(angle)); // Hence the division of those is 1
        }

        static void Numerics()
        {
            // BigInteger
            {
                BigInteger x = new BigInteger(uint.MaxValue); // 2^32 - 1
                BigInteger y = ulong.MaxValue; // 2^64 - 1

                BigInteger z = x * y;
                Console.WriteLine(z);
                Console.WriteLine(BigInteger.Log(z, 2)); // 95.999 ~= 2^(32 + 64)

                BigInteger n = 10000;
                BigInteger fac = 1;
                for (BigInteger i = 2; i < n; i++)
                    fac *= i;

                Console.WriteLine(fac);
            }

            // Complex
            {
                var c1 = new Complex(3.0, 4.0);
                var c2 = Complex.FromPolarCoordinates(1.0, Math.PI / 3 /* 60 degrees */);
                var c3 = Complex.Zero;
                var c4 = Complex.One;
                var c5 = Complex.ImaginaryOne;

                Console.WriteLine(c1); // (3, 4)
                Console.WriteLine(c2); // (0.5, 0.866025403784439) - see Figure 26.14
                Console.WriteLine(c3); // (0, 0)
                Console.WriteLine(c4); // (1, 0)
                Console.WriteLine(c5); // (0, 1)

                Console.WriteLine(c1);
                Console.WriteLine(c2.Real); // Cosine of angle
                Console.WriteLine(c3.Imaginary); // Sine of angle
                Console.WriteLine(c4.Magnitude); // Radius of circle
                Console.WriteLine(c5.Phase); // Angle in radians

                Console.WriteLine(c1 + c2 * c3 - c4 / c5); // Standard arithmetic
                Console.WriteLine(Complex.Pow(c1, c2)); // Refer to theory :-)
                Console.WriteLine(Complex.Reciprocal(c3)); // 1 + 0i divided by c
                Console.WriteLine(Complex.Negate(c4)); // Flip both signs
                Console.WriteLine(Complex.Conjugate(c5)); // Flip imaginary sign
            }
        }

        static void RandomNumbers()
        {
            var r = new Random();

            for (int i = 0; i < 5; i++)
                Console.WriteLine(r.Next(0, 100)); // Within range [0, 100[

            for (int i = 0; i < 5; i++)
                Console.WriteLine(r.NextDouble()); // Always between 0.0 and 1.0

            var bs = new byte[5];
            r.NextBytes(bs); // Fills the given array with bytes
            foreach (byte b in bs)
                Console.WriteLine(b);
        }

        static void ApproximatePi()
        {
            const int TOTAL = 1000000; // Number of shots per turn
            var random = new Random();

            for (int t = 0; t < 10; t++)
            {
                int within = 0;
                for (int i = 0; i < TOTAL; i++)
                {
                    double x = random.NextDouble();
                    double y = random.NextDouble();
                    double r = Math.Sqrt(x * x + y * y);
                    if (r < 1.0) // Within the circle
                        within++;
                }

                var approx = (double)within * 4 / TOTAL;
                Console.WriteLine(approx);
            }
        }

        static void DateTimes()
        {
            // January, 1st of 1 AC at midnight
            var start = new DateTime();
            Console.WriteLine(start);

            // Ticks are measured in 100 nanoseconds, starting from 01/01/2001 12:00:00 AM
            // Hence, the below corresponds to February, 1st of 1 AC at midnight
            var fromTicks = new DateTime(
                10L * 1000 * 1000 * // from 100ns to microsecond, millisecond and second
                60 * 60 * 24 * // from second to minute, hour and day
                31); // 31 days in January
            Console.WriteLine(fromTicks);

            // From year, month and day, at midnight
            var birthday = new DateTime(1983, 2, 11);
            Console.WriteLine(birthday);
            Console.WriteLine(birthday.DayOfWeek); // Friday (a DayOfWeek enum value)
            Console.WriteLine(birthday.DayOfYear); // 42 (31 days in January + 11)

            // Month, day, hour, minute and second are all set to 11
            var elfelfelf = new DateTime(2011, 11, 11, 11, 11, 11);
            Console.WriteLine(elfelfelf);

            // DateTime.Now gets the current time, up to tick precision
            var now = DateTime.Now;
            Console.WriteLine(now);
            Console.WriteLine(now.Year);
            Console.WriteLine(now.Month);
            Console.WriteLine(now.Day);
            Console.WriteLine(now.Hour);
            Console.WriteLine(now.Minute);
            Console.WriteLine(now.Second);
            Console.WriteLine(now.Millisecond);

            // FILETIME interoperability
            long fileTime = now.ToFileTimeUtc();
            DateTime aWhileAgo1 = DateTime.FromFileTimeUtc(fileTime);
            Console.WriteLine(aWhileAgo1);

            // Binary representation
            long binary = now.ToBinary();
            DateTime aWhileAgo2 = DateTime.FromBinary(binary);
            Console.WriteLine(aWhileAgo2);
        }

        static void DateTimeParse()
        {
            // Parse
            {
                var birthday = DateTime.Parse("11/02/1983", new CultureInfo("nl-BE"));
                Console.WriteLine(birthday);
            }

            // ParseExact
            {
                var formats = new[] {
                    "dd/MM/yyyy", "dd/M/yyyy", "d/MM/yyyy", "d/M/yyyy",
                    "dd/MM/yy", "dd/M/yy", "d/MM/yy", "d/M/yy"
                };

                var culture = new CultureInfo("nl-BE");
                var birthday = DateTime.ParseExact("11/2/83", formats, culture, DateTimeStyles.None);
                Console.WriteLine(birthday);
            }
        }

        static void DateTimeToString()
        {
            var birthday = new DateTime(1983, 2, 11);

            // 11/02/1983
            Console.WriteLine(birthday.ToString("dd/MM/yyyy"));

            // 11/02/1983 0:00:00 with 24-hour notation (no AM/PM designator)
            Console.WriteLine(birthday.ToString(new CultureInfo("nl-BE")));

            // Friday, February 11, 1983 - a similar method exists for time
            Console.WriteLine(birthday.ToLongDateString());

            // 2/11/1983 - a similar method exists for time
            Console.WriteLine(birthday.ToShortDateString());
        }

        static void TimeZones()
        {
            var res = from zone in TimeZoneInfo.GetSystemTimeZones()
                      where zone.SupportsDaylightSavingTime
                      select new
                      {
                          zone.Id,
                          zone.DisplayName,
                          Adjustments = zone.GetAdjustmentRules().Length
                      };

            foreach (var zoneInfo in res)
                Console.WriteLine(zoneInfo);

            var brussels = TimeZoneInfo.FindSystemTimeZoneById("Romance Standard Time");
            Console.WriteLine(TimeZoneInfo.ConvertTime(DateTime.Now, brussels));
            Console.WriteLine(TimeZoneInfo.ConvertTimeToUtc(DateTime.Now));
            Console.WriteLine(DateTime.UtcNow);
        }

        static void RelativeTime()
        {
            // TimeSpan values
            {
                var bestTime1 = new TimeSpan(2, 3, 59); // 2 hours, 3 minutes and 59 seconds
                var bestTime2 = TimeSpan.FromHours(2) + TimeSpan.FromMinutes(3) + TimeSpan.FromSeconds(59);

                var hundredDays1 = new TimeSpan(100, 0, 0, 0); // overload with days, h, m, s
                var hundredDays2 = TimeSpan.FromDays(100);
            }

            // Computations with DateTime and TimeSpan
            {
                var startTime = new DateTime(2008, 9, 28, 8, 40, 25);
                var endTime = new DateTime(2008, 9, 28, 10, 44, 24);
                var elapsed = endTime - startTime;
                Console.WriteLine(elapsed);

                var endTime2 = startTime + elapsed;
                Console.WriteLine(endTime == endTime2);

                Console.WriteLine(elapsed.Hours + ":" + elapsed.Minutes);
                Console.WriteLine(elapsed.TotalSeconds);

                var hundredDaysFestival = DateTime.Now.AddDays(100);
                Console.WriteLine(hundredDaysFestival);
            }
        }

        static void Guids()
        {
            // Construct GUID
            var myId = new Guid("5b8a5333-1bfd-494a-b73a-8812d254d89b");
            Console.WriteLine(myId);

            // Generate GUID
            Console.WriteLine(Guid.NewGuid());
        }

        static void Nullability()
        {
            int? x = 42;
            Nullable<int> y = null;

            bool b = x.HasValue; // Same as: bool b = x == null;
            Console.WriteLine(b);

            int c = y.GetValueOrDefault(25); // Similar to: int c = y ?? 25;
            Console.WriteLine(c);

            int d = x.Value; // int d = (int)x;
            Console.WriteLine(d);
        }

        static void Uris()
        {
            var url = new Uri("http://bar:8080/foo/quz?answer=42");
            Console.WriteLine(url.Host); // bar
            Console.WriteLine(url.Port); // 8080
            Console.WriteLine(url.AbsolutePath); // /foo/quz
            Console.WriteLine(url.Query); // ?answer=42

            var builder = new UriBuilder();
            builder.Scheme = Uri.UriSchemeFtp; // or use “ftp”
            builder.Host = "localhost";
            builder.Port = 21; // this is the default and will be omitted
            builder.Path = "downloads";
            builder.UserName = "bartde";
            Console.WriteLine(builder.Uri); // ftp://bartde@localhost/downloads
        }

        static void EnvironmentInteraction()
        {
            Console.WriteLine("Machine name:   " + Environment.MachineName);
            Console.WriteLine("Processor #:    " + Environment.ProcessorCount);
            Console.WriteLine("OS version:     " + Environment.OSVersion);
            Console.WriteLine("64-bit OS:      " + Environment.Is64BitOperatingSystem);
            Console.WriteLine("64-bit process: " + Environment.Is64BitProcess);
            Console.WriteLine("Working set:    " + Environment.WorkingSet / 1024 + "MB");
            Console.WriteLine("CLR version:    " + Environment.Version);
            Console.WriteLine("System32 at:    " + Environment.SystemDirectory);
            Console.WriteLine("Not a service:  " + Environment.UserInteractive);
            Console.WriteLine("User info:      " + Environment.UserDomainName + "\\" + Environment.UserName);
            Console.WriteLine("Boot time:      " + (DateTime.Now - TimeSpan.FromMilliseconds(Environment.TickCount)));

            var proc = Environment.GetEnvironmentVariable("PROCESSOR_IDENTIFIER");
            Console.WriteLine(proc);

            var notepad = Environment.ExpandEnvironmentVariables(@"%WINDIR%\notepad.exe");
            Console.WriteLine(notepad);

            Environment.SetEnvironmentVariable("MSG", "Hello");
            System.Diagnostics.Process.Start("cmd.exe"); // Type echo %MSG% in the command prompt that shows up

            var start = Environment.GetFolderPath(Environment.SpecialFolder.StartMenu);
            Console.WriteLine(start);
        }

        // Rename to Main to use as entry-point
        static void Cmdline(string[] args)
        {
            Console.WriteLine("CommandLine:");
            Console.WriteLine(" " + Environment.CommandLine);

            Console.WriteLine("Main:");
            foreach (var arg in args)
                Console.WriteLine(" " + arg);

            Console.WriteLine("CommandLineArgs:");
            foreach (var arg in Environment.GetCommandLineArgs())
                Console.WriteLine(" " + arg);

            var from = new Uri(Assembly.GetEntryAssembly().CodeBase);
            Console.WriteLine(from.LocalPath);
        }

        static void GcStats()
        {
            int generations = GC.MaxGeneration + 1;
            Console.WriteLine(generations); // 3 for 0 through 2

            while (true)
            {
                long total = GC.GetTotalMemory(false /* don’t trigger collection */);
                Console.WriteLine(total);

                for (int i = 0; i < generations; i++) // most likely all print 0
                    Console.WriteLine("Gen {0} = {1}", i, GC.CollectionCount(i));

                Console.WriteLine("Press ENTER to allocate more");
                Console.ReadLine();

                for (int i = 0; i < 2 * 1024; i++)
                {
                    byte[] b = new byte[1024];
                }
            }
        }

        static Lazy<double> _pi = new Lazy<double>(() =>
        {
            return Math.PI /* long running computation of PI (not really the case here) */;
        });

        static void LazyBeings()
        {
            // Experiment: set a breakpoint in the Lazy<double> lambda expression above
            var radius = 2.0;

            double area = _pi.Value * radius * radius; // Triggers computation of PI
            double circumference = _pi.Value * 2 * radius; // Precomputed value of PI
        }

        static void Tuples()
        {
            var t = new Tuple<int, int, int, int, int, int, int, Tuple<int, int>>(1, 2, 3, 4, 5, 6, 7, new Tuple<int, int>(8, 9));
            Console.WriteLine(t); // (1, 2, 3, 4, 5, 6, 7, 8, 9)
        }

        static void FormattingText()
        {
            Console.WriteLine(123.45.ToString(new CultureInfo("en-US")));
            Console.WriteLine(123.45.ToString(new CultureInfo("nl-BE")));

            var enUS = new CultureInfo("en-US");
            var nlBE = new CultureInfo("nl-BE");
            var jaJP = new CultureInfo("ja-JP");

            // Numbers
            {
                double d = 123.456;
                int i = 1234;
                long l = 123456789;

                // Currency, usable for all numeric values
                // Number of decimal digits determined from culture or explicitly
                Console.WriteLine(d.ToString("C", enUS)); // $ 123.46
                Console.WriteLine(d.ToString("C", nlBE)); // &euro; 123,46
                Console.WriteLine(d.ToString("C", jaJP)); // ¥ 123
                Console.WriteLine(d.ToString("C3", enUS)); // $ 123.456
                Console.WriteLine(d.ToString("C0", nlBE)); // &euro; 123
                Console.WriteLine(d.ToString("C2", jaJP)); // ¥ 123.46

                // Decimal, usable for integral values only
                // Number of digits controls left padding with zeros
                Console.WriteLine(i.ToString("D", enUS)); // 1234
                Console.WriteLine(i.ToString("D5", nlBE)); // 01234

                // Scientific notation with exponent
                // Default decimal precision is 6 but can be controlled
                Console.WriteLine(l.ToString("E", enUS)); // 1.234568E+008
                Console.WriteLine(l.ToString("E3", enUS)); // 1.235E+008

                // Digit placeholders
                // Doesn’t insert zeros if no digit appears at the given position
                Console.WriteLine(i.ToString("#####", enUS)); // 1234

                // Zero placeholders
                // Inserts zeros if no digit appears at the given position
                Console.WriteLine(i.ToString("00000", enUS)); // 01234

                // Decimal point
                // Here mixed used with trailing zero padding
                Console.WriteLine(d.ToString("0###.##", enUS)); // 0123.46
            }

            // DateTime values
            {
                // Date patterns without time
                // Others include date/time, UTC, year/month, roundtrip, etc.
                var bd = new DateTime(1983, 2, 11);
                Console.WriteLine(bd.ToString("d", nlBE)); // 11/02/1983
                Console.WriteLine(bd.ToString("D", nlBE)); // vrijdag 11 februari 1983

                // Custom date/time specifiers
                // ddd = short day of week; dd = day of month; MM = month
                Console.WriteLine(bd.ToString("ddd dd/MM", enUS)); // Fri 11/02
            }

            // Console.WriteLine and String.Format
            {
                Console.WriteLine("{0} costs {1:C}", "Chai", 123.45m);

                var description = string.Format(nlBE, "{0} costs {1:C}", "Chai", 123.45m);
                Console.WriteLine(description);

                var musicGroup = string.Format("{1}{0}{0}{1}", "B", 'A'); // Returns ABBA
                Console.WriteLine(musicGroup);

                var products = new[] {
                    new { Name = "Chai", Price = 1.25m },
                    new { Name = "Expensive lawn mower", Price = 495.95m },
                    new { Name = "Chang", Price = 5m },
                };

                Console.WriteLine("{0,-30} {1,8:C}", "Product name", "Price");
                Console.WriteLine("{0,-30} {1,8:C}", new string('-', 30), new string('-', 8));

                foreach (var product in products)
                    Console.WriteLine("{0,-30} {1,8:C}", product.Name, product.Price);
            }
        }

        static void ParsingText()
        {
            double value = 1.0 / 3.0;
            string text = value.ToString();
            Console.WriteLine(text); // 0.333333333333333

            double res1 = double.Parse(text);
            Console.WriteLine(res1);
            Console.WriteLine(value == res1); // False

            string precise = value.ToString("R");
            Console.WriteLine(precise); // 0.33333333333333331

            double res2 = double.Parse(precise);
            Console.WriteLine(res2);
            Console.WriteLine(value == res2); // True
        }

        static void RegularExpressions1()
        {
            var regex = new Regex("0|([1-9][0-9]*)");

            while (true)
            {
                Console.Write("> ");
                string input = Console.ReadLine();
                Match match = regex.Match(input);
                if (match.Success)
                    Console.WriteLine(match.Value);
            }
        }

        static void RegularExpressions2()
        {
            var regex = new Regex("0|([1-9][0-9]*)");

            while (true)
            {
                Console.Write("> ");
                string input = Console.ReadLine();
                foreach (Match match in regex.Matches(input))
                    Console.WriteLine("{0} @ {1}", match.Value, match.Index);
            }
        }

        static void RegularExpressionGroups()
        {
            var regex = new Regex(@"(?<from>0|([1-9][0-9]*))\-(?<to>0|([1-9][0-9]*))");

            while (true)
            {
                Console.Write("> ");
                string input = Console.ReadLine();
                Match match = regex.Match(input);
                if (match.Success)
                    Console.WriteLine("from {0} to {1}", match.Groups["from"].Value, match.Groups["to"].Value);
            }
        }

        static void StringMethods()
        {
            // Predicates
            {
                var s = "Bart";
                Console.WriteLine(s.StartsWith("B")); // ~ s matches B*
                Console.WriteLine(s.EndsWith("t"));   // ~ s matches *t
                Console.WriteLine(s.Contains("a"));   // ~ s matches *a*
            }

            // Mutation (or not)
            {
                string s = "world.";
                s.Insert(0, "Hello ");
                s.Replace('.', '!');
                Console.WriteLine(s); // Contrary to what you may expect, this prints “world.”
            }

            // Creation of new strings...
            {
                string s = "world.";
                s = s.Insert(0, "Hello ");
                s = s.Replace('.', '!');
                Console.WriteLine(s); // Prints “Hello world!”
            }

            // Trimming and padding
            {
                Console.Write("Enter your name: ");
                string input = Console.ReadLine();
                string name = input.Trim(' ', '\t');
                Console.WriteLine("Welcome {0}!", name);

                string price = "$123.45";
                Console.WriteLine(price.PadLeft(10)); // Prints “   $123.45”
                Console.WriteLine("{0,10}");          // Prints “   $123.45” as well
            }

            // Splitting and joining
            {
                Console.Write("Enter a comma-separated list of entries: ");
                string input = Console.ReadLine();
                string[] parts = input.Split(',');
                string result = string.Join("; ", parts);
                Console.WriteLine(result);
            }
        }

        static void StringBuilders()
        {
            var people = new[] {
                new { Name = "Bart", Age = 28 },
                new { Name = "John", Age = 62 },
            };

            var sb = new StringBuilder();
            foreach (var person in people)
                sb.AppendFormat("{0}\t{1}\r\n", person.Name, person.Age);
            Console.WriteLine(sb.ToString());
        }

        static void StringEncoding()
        {
            var bytes = Encoding.ASCII.GetBytes("Hello ASCII!");
            var hello = Encoding.ASCII.GetString(bytes);
            Console.WriteLine(hello);
        }
    }
}
