﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.Contracts;
using System.IO;
using System.Linq;
using System.Threading;

namespace Chapter27
{
    class Program
    {
        static void Main()
        {
            // Call a sample method here.
        }

        static void CodeAnalysis()
        {
            // Produces a warning due to enabling Code Analysis.
            var xs = new int[100, 100];
            for (int i = 0; i < 100; i++)
                for (int j = 0; j < 100; j++)
                    xs[i, j] = 42;
        }

        static void Asserts()
        {
            var xs = new[] { 7, 3, 5, 2, 8, 1, 0 };
            xs.Sort();

            foreach (var x in xs)
                Console.WriteLine(x);
        }

        static void Contracts()
        {
            Console.WriteLine(new MyCollection().Count);
        }

        static void DebuggerOutput()
        {
            ContactPaymentService("Bart", 123.45m);
        }

        static void ContactPaymentService(string account, decimal amount)
        {
            Debug.WriteLine("Connecting to payment service");
            try
            {
                // More code goes here...
                // Assume it throws :-(
                throw new IOException("Oops!");
            }
            catch (IOException ex)
            {
                Debug.WriteLine("Failure to communicate: " + ex.Message);
                // Production code to handle the exception...
            }
        }

        static void ConditionalAttribute()
        {
            // Try running in Debug and Release builds.
            int i = 0;
            Bar(i++);
            Console.WriteLine(i);
        }

        [Conditional("DEBUG")]
        static void Bar(int i) { }

        static void Bar2()
        {
            Debug.Indent();
            Debug.WriteLine("Bar - Begin");

            // Code goes here

            Debug.WriteLine("Bar - End");
            Debug.Unindent();
        }

        static void ControllingTheDebugger()
        {
            Debugger.Break();
            // More code goes here...
        }

        static void StackTraces()
        {
            Bar3();
        }

        static void Bar3()
        {
            var trace = new StackTrace();
            Console.WriteLine(trace);
        }

        static void MeasuringPerformance_Wrong()
        {
            for (int j = 0; j < 10; j++)
            {
                var start = DateTime.Now;

                var n = 1L;
                for (int i = 2; i < 50; i++)
                    n *= i;

                var end = DateTime.Now;
                Console.WriteLine(end - start);
            }
        }

        static void MeasuringPerformance_Good()
        {
            for (int j = 0; j < 10; j++)
            {
                var sw = new Stopwatch();
                sw.Start();

                var n = 1L;
                for (int i = 2; i < 50; i++)
                    n *= i;

                sw.Stop();
                Console.WriteLine(sw.Elapsed);
            }
        }

        static void EventLogs()
        {
            const string SOURCE = "Chapter27";
            const string APPLICATION = "Application";

            if (!EventLog.SourceExists(SOURCE))
                EventLog.CreateEventSource(SOURCE, APPLICATION);

            const int STARTED = 1001;
            const short LIFECYCLE = 1;

            using (var log = new EventLog(APPLICATION, ".", SOURCE))
            {
                log.WriteEntry("Application started.", EventLogEntryType.Information, STARTED, LIFECYCLE);
            }
        }

        static void PerformanceCounters()
        {
            const string SOURCE = "Chapter27";
            const string DESCRIPTION = "Sample counters";

            const string TXCOUNT = "# of transactions";
            const string PROCRATE = "Processing rate";
            const string PROCRATEBASE = "Processing rate base";

            if (!PerformanceCounterCategory.Exists(SOURCE))
            {
                PerformanceCounterCategory.Create(SOURCE, DESCRIPTION,
                    PerformanceCounterCategoryType.SingleInstance,
                    new CounterCreationDataCollection
                    {
                        new CounterCreationData(TXCOUNT, "Keeps track of number of bank transfers", PerformanceCounterType.NumberOfItems32),
                        new CounterCreationData(PROCRATE, "Average transaction processing rate", PerformanceCounterType.AverageCount64),
                        new CounterCreationData(PROCRATEBASE, "Denominator for processing rate counter", PerformanceCounterType.AverageBase),
                    }
                );
            }

            var txCount = new PerformanceCounter(SOURCE, TXCOUNT, false);
            var procRate = new PerformanceCounter(SOURCE, PROCRATE, false);
            var procRateBase = new PerformanceCounter(SOURCE, PROCRATEBASE, false);

            Console.WriteLine("Press ENTER to start processing...");
            Console.ReadLine();
            var rand = new Random();
            var queue = new ConcurrentQueue<int>();
            new Thread(() =>
            {
                while (true)
                {
                    Thread.Sleep(rand.Next(1000, 5000));

                    int n = 0;
                    int tx;
                    while (queue.TryDequeue(out tx))
                    {
                        Console.WriteLine("Processing transaction {0}", tx);
                        Thread.Sleep(rand.Next(0, 100));
                        txCount.Decrement();
                        n++;
                    }

                    procRate.IncrementBy(n);
                    procRateBase.Increment();
                }
            }).Start();

            for (int tx = 0; true; tx++)
            {
                Console.WriteLine("Requesting transaction {0}", tx);
                queue.Enqueue(tx);
                txCount.Increment();

                Thread.Sleep(rand.Next(0, 200));
            }
        }

        static void Processes_Querying()
        {
            var memoryHogs = from p in Process.GetProcesses()
                             where p.WorkingSet64 > 50 * 1024 * 1024 /* 50 MB */
                             select p;

            foreach (var memoryHog in memoryHogs)
                Console.WriteLine(memoryHog);
        }

        static void Processes_Creation()
        {
            var cmd = Process.Start("cmd.exe");
            Console.WriteLine("We’re not blocked while cmd.exe is running...");
            cmd.WaitForExit();
            Console.WriteLine(cmd.ExitCode);
        }

        static void Processes_MonitoringExit()
        {
            var cmd = new Process
            {
                EnableRaisingEvents = true,
                StartInfo = new ProcessStartInfo("cmd.exe")
            };
            cmd.Exited += (o, e) => Console.WriteLine(cmd.ExitCode);
            cmd.Start();

            while (true)
            {
                Thread.Sleep(500);
                Console.Write(".");
            }
        }

        static void Processes_IO()
        {
            var tlist = new Process
            {
                StartInfo = new ProcessStartInfo("tasklist.exe")
                {
                    RedirectStandardOutput = true,
                    UseShellExecute = false
                }
            };
            tlist.Start();
            Console.WriteLine(tlist.StandardOutput.ReadToEnd());
        }
    }

    public static class Ext
    {
        public static void Sort<T>(this T[] values)
        {
            if (values == null)
                throw new ArgumentNullException("values");

            Sort(values, 0, values.Length - 1);
        }

        private static void Sort<T>(T[] values, int from, int to)
        {
            Debug.Assert(from <= to, "Expected from to be less than to.");

            #region Implementation left as an exercise
            // Left as an exercise; incorrect recursive call shown below, to trigger the assert.
            Sort(values, 10, 9);
            #endregion
        }
    }

    public class MyCollection : IEnumerable<int>
    {
        public int Count
        {
            get
            {
                Contract.Ensures(Contract.Result<int>() >= 0);
                return 10; // Try returning a negative value...
            }
        }

        public IEnumerator<int> GetEnumerator()
        {
            throw new NotImplementedException();
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            throw new NotImplementedException();
        }
    }
}
