﻿using System;
using System.IO;
using System.IO.Compression;
using System.IO.MemoryMappedFiles;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;

namespace Chapter28
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void ListingDrives()
        {
            foreach (var drive in DriveInfo.GetDrives())
            {
                Console.WriteLine("{0} - {1}", drive.Name, drive.DriveType);
                if (drive.IsReady)
                {
                    Console.WriteLine(drive.DriveFormat);
                    Console.WriteLine(string.IsNullOrEmpty(drive.VolumeLabel) ? "(no label)" : drive.VolumeLabel);
                    Console.WriteLine("{0} free of {1}", GetSize(drive.TotalFreeSpace), GetSize(drive.TotalSize));
                }
                else
                {
                    Console.WriteLine("(No disk present)");
                }

                Console.WriteLine();
            }
        }

        static string GetSize(double size)
        {
            var prefixes = new[] { "bytes", "KB", "MB", "GB", "TB", "PB", "EB" };

            int i;
            for (i = 0; size > 1000; i++)
            {
                size /= 1024;
            }

            return string.Format("{0:##0.##} {1}", size, prefixes[i]);
        }

        static void Directories()
        {
            Console.Write("Enter path: ");
            var path = Console.ReadLine();
            Console.WriteLine();

            var folder = new DirectoryInfo(path);
            if (folder.Exists)
            {
                Console.WriteLine("Full name: " + folder.FullName);
                Console.WriteLine("Name: " + folder.Name);
                Console.WriteLine("Created: " + folder.CreationTime);
                Console.WriteLine("Parent: " + folder.Parent.FullName);
                Console.WriteLine("Root: " + folder.Root);

                foreach (var child in folder.GetFileSystemInfos())
                    Console.WriteLine("{0} {1}", child is FileInfo ? " " : "D", child);

                foreach (var exe in folder.EnumerateFiles("*.exe", SearchOption.AllDirectories))
                    Console.WriteLine(exe.FullName);
            }

            if (!folder.Exists)
                folder.Create();

            var subfolder = folder.CreateSubdirectory("Bar");
            Console.WriteLine(subfolder.CreationTime);

            subfolder.Delete(/* can specify to recurse into subdirectories */);
            Console.WriteLine(subfolder.Exists);
            subfolder.Refresh();
            Console.WriteLine(subfolder.Exists);
        }

        static void Paths()
        {
            var path = @"c:\temp\test.txt";

            Console.WriteLine("Absolute: " + Path.IsPathRooted(path));
            Console.WriteLine("Root: " + Path.GetPathRoot(path));
            Console.WriteLine("Directory: " + Path.GetDirectoryName(path));

            Console.WriteLine("File name: " + Path.GetFileName(path));
            if (Path.HasExtension(path))
                Console.WriteLine("Extension: " + Path.GetExtension(path));

            Console.WriteLine("Full path: " + Path.GetFullPath("test.txt"));

            Console.WriteLine(Path.GetTempPath());
            Console.WriteLine(Path.GetRandomFileName());

            // Creates an empty file on disk under the %TEMP% location.
            Console.WriteLine(Path.GetTempFileName());
        }

        static void Files()
        {
            Console.Write("Enter path: ");
            var path = Console.ReadLine();
            Console.WriteLine();

            var file = new FileInfo(path);
            if (file.Exists)
            {
                Console.WriteLine("Created: " + file.CreationTime);
                Console.WriteLine("Accessed: " + file.LastAccessTime); // Bogus...
                Console.WriteLine("Modified: " + file.LastWriteTime);
                Console.WriteLine("Directory: " + file.Directory.FullName);
                Console.WriteLine("Read-only: " + file.IsReadOnly);
                Console.WriteLine("Attributes: " + file.Attributes);
                Console.WriteLine("Size: " + GetSize(file.Length));
            }
        }

        static void MonitoringFileSystemActivity()
        {
            using (var watcher = new FileSystemWatcher(Path.GetTempPath(), "*.*"))
            {
                watcher.Created += (o, e) =>
                {
                    Console.WriteLine("Created: " + e.FullPath);
                };

                watcher.EnableRaisingEvents = true;

                Console.WriteLine("Monitoring... Press ENTER to exit.");
                Console.Read();
            }
        }

        static void FileMethods()
        {
            string[] lines = File.ReadAllLines(@"..\..\Program.cs");
            foreach (var line in lines)
                Console.WriteLine(line);

            File.WriteAllLines(
                @"..\..\Program.bak",
                from line in File.ReadLines(@"..\..\Program.cs")
                where !string.IsNullOrWhiteSpace(line)
                select line.TrimStart()
            );
        }

        static void WritersAndReaders()
        {
            // Writing text
            using (StreamWriter sw = File.CreateText("log.txt"))
            {
                sw.AutoFlush = true;

                sw.WriteLine("{0} - Started", DateTime.Now);

                // Reading line by line
                using (StreamReader sr = File.OpenText(@"..\..\Program.cs"))
                {
                    while (!sr.EndOfStream)
                        Console.WriteLine(sr.ReadLine());
                }

                // Reading in blocks
                using (StreamReader sr = File.OpenText(@"..\..\Program.cs"))
                {
                    var buffer = new char[1024];
                    int read = sr.Read(buffer, 0, buffer.Length);
                    Console.WriteLine("{0} characters read", read);
                }

                sw.WriteLine("{0} - Stopped", DateTime.Now);
            }

            // StringWriter and StringReader
            {
                StringBuilder sb = new StringBuilder();
                TextWriter writer = new StringWriter(sb);

                string input = Console.ReadLine();
                writer.WriteLine(input);

                TextReader reader = new StringReader(sb.ToString());
                string line = reader.ReadLine();
                Console.WriteLine(line);
            }
        }

        static void MemoryStreams()
        {
            var bs = new byte[20];
            var ms1 = new MemoryStream(bs, index: 5, count: 10, writable: true);
            var ms2 = new MemoryStream(bs, index: 10, count: 10, writable: false);

            ms1.Seek(5, SeekOrigin.Begin);     //      @ relative = 5, absolute = 15
            ms1.Write(new byte[] { 9, 8, 7 }, 0, 3);

            Console.WriteLine(bs[10]);         // 0x09 @               absolute = 10
            Console.WriteLine(ms2.ReadByte()); // 0x09 @ relative = 0, absolute = 10
            Console.WriteLine(ms2.ReadByte()); // 0x08 @ relative = 1, absolute = 11
            Console.WriteLine(ms2.ReadByte()); // 0x07 @ relative = 2, absolute = 12

            bs[18] = 1;                        // 0x01 @               absolute = 18
            ms2.Seek(5, SeekOrigin.Current);
            Console.WriteLine(ms2.ReadByte()); // 0x01 @ relative = 8, absolute = 18
        }

        static void WorkingWithFiles()
        {
            // FileStream Read methods
            using (FileStream fs = File.OpenRead(@"c:\windows\system32\notepad.exe"))
            {
                var m = (char)fs.ReadByte();
                var z = (char)fs.ReadByte();
                Console.WriteLine("{0}{1}", m, z);
            }

            // BinaryReader usage
            using (FileStream fs = File.OpenRead(@"c:\windows\system32\notepad.exe"))
            using (var br = new BinaryReader(fs))
            {
                br.BaseStream.Position = 0x3c;
                int offset = br.ReadInt32();

                br.BaseStream.Position = offset;
                char[] signature = br.ReadChars(2);
                Console.WriteLine(new string(signature));
            }
        }

        static void AsynchronousFileAccess()
        {
            using (FileStream fs = File.OpenRead(@"c:\windows\system32\notepad.exe"))
            {
                var buffer = new byte[1024 * 1024]; // 1 MB buffer
                IAsyncResult iar = fs.BeginRead(buffer, 0, buffer.Length, result =>
                {
                    // This code will be called upon completion.
                    int count = fs.EndRead(result);
                    Console.WriteLine("{0} bytes read", count);
                }, null);

                Console.Write("Read operation started");
                while (!iar.IsCompleted)
                {
                    Console.Write("."); // Here we could be doing other useful work.
                }

                Console.ReadLine();
            }
        }

        static void AsynchronousDelegateInvocation()
        {
            Func<int, int> twice = x => x * 2;

            int n = 21;
            twice.BeginInvoke(n, iar =>
            {
                int res = twice.EndInvoke(iar);
                Console.WriteLine("And the answer to ({0} * 2) is {1}!", n, res);
            }, null /* the dreaded state parameter */);
        }

        static void Zipping()
        {
            const string SRC = @"..\..\Program.cs";
            const string TGT = @"..\..\Program.gz";

            using (var zipped = File.Create(TGT))
            using (var compress = new GZipStream(zipped, CompressionMode.Compress))
            using (var source = File.OpenRead(SRC))
            {
                source.CopyTo(compress);
            }

            Console.WriteLine("Source length = " + new FileInfo(SRC).Length);
            Console.WriteLine("Zipped length = " + new FileInfo(TGT).Length);
        }

        static void MemoryMappedIO()
        {
            using (var file = MemoryMappedFile.CreateFromFile(@"c:\windows\Notepad.exe"))
            {
                using (var accessor = file.CreateViewAccessor())
                {
                    int offset = accessor.ReadInt32(0x3c);
                    Header header;
                    accessor.Read<Header>(offset + 4, out header);
                    // Print header fields of interest
                }
            }
        }
    }

    [StructLayout(LayoutKind.Explicit)]
    struct Header
    {
        [FieldOffset(0)]
        public Machine Machine;

        [FieldOffset(2)]
        public ushort NumberOfSections;

        [FieldOffset(18)]
        public Characteristics Characteristics;
    }

    enum Machine : ushort
    {
        Amd64 = 0x8664,
        I386 = 0x14c,
        IA64 = 0x200,
        // Most other documented machine architectures are extinct (e.g. MIPS16)
    }

    [Flags]
    enum Characteristics : ushort
    {
        ExecutableImage = 0x2,
        LargeAddressAware = 0x20,
        Machine32Bit = 0x100,
        // Lots of other flags exist, see specification
    }
}
