﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.IO;
using System.Diagnostics;

namespace Chapter29
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void CreatingThreads()
        {
            var worker = new Thread(() =>
            {
                Console.WriteLine("Worker thread - Hello!");
            });

            Console.WriteLine("Main thread - Press ENTER to start worker");
            Console.ReadLine();
            worker.Start();

            Console.WriteLine("Main thread - Worker started");
            Console.WriteLine("Main thread - Press ENTER to quit");
            Console.ReadLine();
        }

        static void ParameterizedThreads()
        {
            var worker = new Thread(message =>
            {
                Console.WriteLine("Worker thread - {0}", message);
            });

            Console.Write("Main thread - Enter message to send to worker: ");
            var input = Console.ReadLine();
            worker.Start(input);
        }

        static void StoppingThreads()
        {
            var stopped = false;

            var clock = new Thread(() =>
            {
                while (!stopped)
                {
                    Console.WriteLine("Tick - " + DateTime.Now);
                    Thread.Sleep(1000);
                }
            });

            clock.Start();

            Console.WriteLine("Press ENTER to stop the clock...");
            Console.ReadLine();

            stopped = true;
            clock.Join();
        }

        static void BackgroundVersusForeground()
        {
            var t1 = new Thread(() =>
            {
                for (int i = 0; i < 10; i++)
                {
                    Console.WriteLine("Background says {0}", i);
                    Thread.Sleep(500);
                }
            }) { IsBackground = true };

            var t2 = new Thread(() =>
            {
                for (int i = 0; i < 10; i++)
                {
                    Console.WriteLine("Foreground says {0}", i);
                    Thread.Sleep(100);
                }
            });

            t1.Start();
            t2.Start();
        }

        static void NamingThreads()
        {
            var monitor1 = new Thread(() =>
            {
                while (true)
                {
                    Console.WriteLine("Monitor 1 heartbeat");
                    // Do something more useful every second...
                    Thread.Sleep(1000);
                }
            }) { IsBackground = true, Name = "Monitoring thread 1" };

            // ...
            monitor1.Start();

            // Prevent the main thread from exiting for the sample...
            Console.ReadLine();
        }

        static void ThreadExceptions()
        {
            // This will be fatal!
            new Thread(() =>
            {
                Thread.Sleep(1000);
                throw new InvalidOperationException("Oops!");
            }).Start();

            while (true)
            {
                Console.Write(".");
                Thread.Sleep(100);
            }
        }

        [ThreadStatic]
        static int s_i;

        static void ThreadSpecificState()
        {
            new Thread(() =>
            {
                while (true)
                {
                    s_i++;
                }
            }).Start();

            while (true)
            {
                Thread.Sleep(1000);
                Console.WriteLine(s_i);
            }
        }

        [ThreadStatic]
        static TextWriter s_logger;

        static void ThreadSpecificState2()
        {
            new Thread(() =>
            {
                EnsureLog();
                // Some thread that’s responsible for some I/O operation.
                // At various times, it writes to the log, as shown here:
                s_logger.WriteLine("{0} - Performed some action", DateTime.Now);
                // Much more code is executed on this thread...
            }) { Name = "Lazy writer" }
            .Start();

            // Others threads could be created, each of which has to call
            // the EnsureLog method to have its s_logger field set up properly.
        }

        static void EnsureLog()
        {
            string name = Thread.CurrentThread.Name ?? Thread.CurrentThread.ManagedThreadId.ToString();

            // More complicated initialization logic could be dreamed up.
            // What matters is the fact this method will run multiple times
            // on different threads, initializing that thread’s logger.
            s_logger = File.CreateText(name + ".log");
        }

        static ThreadLocal<TextWriter> s_log = new ThreadLocal<TextWriter>(() =>
        {
            string name = Thread.CurrentThread.Name ?? Thread.CurrentThread.ManagedThreadId.ToString();

            // More complicated initialization logic could be dreamed up.
            // What matters is the fact this method will run multiple times
            // on different threads, initializing that thread’s logger.
            return File.CreateText(name + ".log");
        });

        static void ThreadSpecificState_ThreadLocal()
        {
            new Thread(() =>
            {
                // Some thread that’s responsible for some I/O operation.
                // At various times, it writes to the log, as shown here:
                s_log.Value.WriteLine("{0} - Performed some action", DateTime.Now);
                // Much more code is executed on this thread...
            }) { Name = "Lazy writer" }
            .Start();

            // Others threads could be created
            Console.ReadLine();
        }

        static void ThreadLocalInClosure()
        {
            var i = new ThreadLocal<int>();
            new Thread(() =>
            {
                while (true)
                    i.Value++;
            }).Start();

            // Always prints 0
            while (true)
            {
                Thread.Sleep(1000);
                Console.WriteLine(i.Value);
            }
        }

        static void DebuggingExercise()
        {
            // Background thread.
            new Thread(() =>
            {
                while (true)
                {
                    Thread.Sleep(30);
                    Console.WriteLine("{0} - {1:hh:mm:ss:fff}", Thread.CurrentThread.ManagedThreadId, DateTime.Now);
                }
            }).Start();

            // Main thread’s code.
            while (true)
            {
                Thread.Sleep(100);
                Console.WriteLine("{0} - {1:hh:mm:ss:fff}", Thread.CurrentThread.ManagedThreadId, DateTime.Now);
            }
        }

        static void ThreadPools()
        {
            Console.Write("Press ENTER to start... ");
            Console.ReadLine();

            Console.WriteLine("Scheduling work...");
            for (int i = 0; i < 1000; i++)
            {
                ThreadPool.QueueUserWorkItem(_ =>
                {
                    Thread.Sleep(100);
                });
            }

            Console.ReadLine();
        }

        static void ThreadPools_PerformanceCharacteristics()
        {
            const int N = 1000;
            var evt = new CountdownEvent(N);

            Console.WriteLine("Scheduling work...");

            var sw = new Stopwatch();
            sw.Start();

            for (int i = 0; i < N; i++)
            {
                new Thread(_ =>
                {
                    Thread.Sleep(100);
                    evt.Signal();
                }).Start();
            }

            evt.Wait();
            sw.Stop();

            Console.WriteLine(sw.ElapsedMilliseconds);
        }

        static void ThreadPools_BewareOfThreadLocalState()
        {
            var state = new ThreadLocal<int>();
            for (int i = 0; i < 100; i++)
            {
                ThreadPool.QueueUserWorkItem(_ =>
                {
                    Console.WriteLine("{0} @ {1}", state.Value++, Thread.CurrentThread.ManagedThreadId
                    );
                });
            }

            Console.ReadLine();
        }

        static void Synchronization_Missing()
        {
            int n = 0;

            var up = new Thread(() =>
            {
                for (int i = 0; i < 1000000; i++)
                    n++;
            });

            up.Start();

            for (int i = 0; i < 1000000; i++)
                n--;

            up.Join();
            Console.WriteLine(n);
        }

        static void Synchronization_Lock()
        {
            object gate = new object();
            int n = 0;

            var up = new Thread(() =>
            {
                for (int i = 0; i < 1000000; i++)
                    lock (gate)
                        n++;
            });

            up.Start();

            for (int i = 0; i < 1000000; i++)
                lock (gate)
                    n--;

            up.Join();
            Console.WriteLine(n);
        }

        static void Synchronization_Mutex()
        {
            var mutex = new Mutex();
            int n = 0;

            var up = new Thread(() =>
            {
                for (int i = 0; i < 1000000; i++)
                {
                    mutex.WaitOne();
                    n++;
                    mutex.ReleaseMutex();
                }
            });

            up.Start();

            for (int i = 0; i < 1000000; i++)
            {
                mutex.WaitOne();
                n--;
                mutex.ReleaseMutex();
            }

            up.Join();
            Console.WriteLine(n);
        }

        static void MutexFun(string[] args)
        {
            // Make this the Main method to run the sample shown in the book.
            var createdNew = false;
            var mutex = new Mutex(/* start as owned */ false, args[0], out createdNew);
            Console.WriteLine("Mutex {0} - Created = {1}", args[0], createdNew);

            Console.Write("Waiting... ");
            mutex.WaitOne();
            Console.WriteLine("Acquired.", DateTime.Now);

            Console.Write("Press ENTER to release... ");
            Console.ReadLine();
            mutex.ReleaseMutex();
        }

        static void Semaphores()
        {
            var rnd = new Random();
            var railwaySemaphore = new Semaphore(2, 2);
            for (int i = 0; i < 10; i++)
            {
                new Thread(() =>
                {
                    // Time to leave the railway station.
                    Thread.Sleep(rnd.Next(500, 8000));
                    Console.WriteLine(Thread.CurrentThread.Name + " departing.");

                    // Time to reach the semaphore next to the track.
                    Thread.Sleep(rnd.Next(100, 5000));
                    Console.WriteLine(Thread.CurrentThread.Name + " at signal.");

                    // Waiting for the first signal.
                    // The train is allowed to pass if no two trains are in the segment.
                    railwaySemaphore.WaitOne();
                    Console.WriteLine(Thread.CurrentThread.Name + " passed signal.");

                    // Time to reach the end of the protected railway segment.
                    Thread.Sleep(rnd.Next(2000, 4000));
                    Console.WriteLine(Thread.CurrentThread.Name + " exited segment.");
                    railwaySemaphore.Release();
                }) { Name = "Train " + i }.Start();
            }
        }

        static void ReaderWriterLocks()
        {
            var @lock = new ReaderWriterLockSlim();

            int counter = 0;
            new Thread(() =>
            {
                while (true)
                {
                    @lock.EnterWriteLock();
                    counter++;
                    counter--;
                    @lock.ExitWriteLock();
                }
            }).Start();

            while (true)
            {
                Thread.Sleep(100);

                @lock.EnterReadLock();
                Console.WriteLine(counter);
                @lock.ExitReadLock();
            }
        }

        static void OrchestraUsingEvents()
        {
            var conductorSignalsViolins = new ManualResetEvent(false);
            var conductorSignalsTrumpetists = new ManualResetEvent(false);

            for (int i = 0; i < 5; i++)
            {
                new Thread(() =>
                {
                    conductorSignalsViolins.WaitOne();
                    for (int j = 0; j < 20; j++)
                    {
                        Console.Write(Thread.CurrentThread.Name + " ");
                        Thread.Sleep(100);
                    }
                }) { Name = "V" + i }.Start();
            }

            for (int i = 0; i < 3; i++)
            {
                new Thread(() =>
                {
                    conductorSignalsTrumpetists.WaitOne();
                    for (int j = 0; j < 20; j++)
                    {
                        Console.Write(Thread.CurrentThread.Name + " ");
                        Thread.Sleep(150);
                    }
                }) { Name = "T" + i }.Start();
            }

            Thread.Sleep(1000); // Time for the trumpets
            conductorSignalsTrumpetists.Set();

            Thread.Sleep(2000); // Time for the violins
            conductorSignalsViolins.Set();
        }

        static void OrchestraUsingCountdown()
        {
            var start = new CountdownEvent(8);
            var conductorSignalsViolins = new ManualResetEvent(false);
            var conductorSignalsTrumpetists = new ManualResetEvent(false);

            for (int i = 0; i < 5; i++)
            {
                new Thread(() =>
                {
                    Console.WriteLine(Thread.CurrentThread.Name + " takes a seat.");
                    start.Signal();
                    conductorSignalsViolins.WaitOne();
                    // Code to play my part...
                }) { Name = "V" + i }.Start();
            }

            for (int i = 0; i < 3; i++)
            {
                new Thread(() =>
                {
                    Console.WriteLine(Thread.CurrentThread.Name + " takes a seat.");
                    start.Signal();
                    conductorSignalsTrumpetists.WaitOne();
                    // Code to play my part...
                }) { Name = "T" + i }.Start();
            }

            start.Wait();
            // Orchestrator logic to set various events...
        }

        static void OrchestraWaitAny()
        {
            var conductorSignalsViolins = new ManualResetEvent(false);
            var emergencyAlarm = new ManualResetEvent(false);

            for (int i = 0; i < 5; i++)
            {
                new Thread(() =>
                {
                    int res = WaitHandle.WaitAny(new[] { conductorSignalsViolins, emergencyAlarm });
                    if (res == 0)
                    {
                        // Code to play my part...
                    }
                    else
                        // Just quitting from the thread would suffice, but let’s be
                        // a bit more illustrative here...
                        Environment.FailFast("Run for your life!");
                }) { Name = "V" + i }.Start();
            }

            // The orchestra conductor could signal to start playing - or -
            // someone could set off the fire alarm.
        }

        static void InterlockedOperations()
        {
            int n = 0;

            var up = new Thread(() =>
            {
                for (int i = 0; i < 1000000; i++)
                    Interlocked.Increment(ref n);
            });

            up.Start();

            for (int i = 0; i < 1000000; i++)
                Interlocked.Decrement(ref n);
            
            up.Join();
            Console.WriteLine(n);
        }

        static void BackgroundWorkerSample()
        {
            new FileCopier().ShowDialog();
        }
    }
}
